# -*- coding: utf-8 -*-
from .sentence_processor import SentenceProcessor,sentence_convert_examples_to_features,sigmoid_array

from .aspect_processor import AspectProcessor,aspect_convert_examples_to_features,tag2ot,tag2ts,evaluate_ote,evaluate_ts,ASPECT_LABEL_MAP,SENTIMENT_LABEL_MAP,chg_labels_to_aspect_and_sentiment

from .general import split_text_from_input_data,get_toolkit_models,get_domain_embedding


# coding=utf-8
# Some of structures or codes in this script are referenced from HuggingFace Inc. team. 

# Copyright 2018 The Google AI Language Team Authors and The HuggingFace Inc. team.
# Copyright (c) 2018, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
'''句子情感分類的預測相關程式'''

import logging
import os
import sys
import tempfile
import numpy as np
import pandas as pd
import torch
from torch.utils.data import DataLoader, SequentialSampler, TensorDataset , DistributedSampler
from tqdm import tqdm
from pathlib import Path

from transformers import (
    BertTokenizer,  
)

from .model_structure import MyNewBertForSequenceClassification  
from .utils import SentenceProcessor,sentence_convert_examples_to_features as convert_examples_to_features,sigmoid_array,split_text_from_input_data,get_toolkit_models

logger = logging.getLogger(__name__) 

SENTENCE_CLASSIFICATION_MODEL = {
   'default' : '/pretrained_model/chinese_sentence_model',  #以本研究繼續預訓練模型進行微調後的最佳模型
   'open' : '/pretrained_model/chinese_sentence_model'      #以開源預訓練模型進行微調後的最佳模型
} 
  
DEFAULT_SENTENCE_CLASSIFICATION_TOKENZIER = 'bert-base-chinese' 


class SentenceSentimentClassification:
    '''句子情感分類的主要類別：使用提供的微調模型來預測給定句子的情感'''  
    
    def __init__(self,
                 model_name_or_path = 'default',
                 tokenizer = DEFAULT_SENTENCE_CLASSIFICATION_TOKENZIER,
                 model_cache_folder_name = 'original',
                 no_cuda = False,
                 local_rank = -1,
                 logging_level = "info"
                ):  
        '''
        設定句子情感分類相關參數：
        model_name_or_path : 用來預測句子情感的模型，可以為放置模型的路徑、或是此工具所提供的模型名稱；默認為本工具提供的由繼續預訓練模型微調完畢的模型(參數值為"default")，如果想使用由開源預訓練模型所微調出的模型，參數值設為"open"
        tokenizer : 模型使用的tokenizer名稱，默認為"bert-base-chinese" 
        model_cache_folder_name : 存放本研究所開源模型的資料夾名稱，預設名稱為"original"，表示模型下載後存放於/pretrained_model/chinese_sentence_model/original資料夾中 
        no_cuda : 是否避免使用gpu，預設"False"
        local_rank : 是否使用平行化運算，預設"-1"
        logging_level : 設置logging level，預設為"info"
        '''
        
        
        self.local_rank = local_rank
        
        llevel = logging.WARN
        self.logging_display = False
        if logging_level == "info":
            llevel = logging.INFO
            self.logging_display = True
            
        
        ### 設定 logging ###
        
        logging.basicConfig(
            format = "%(asctime)s - %(levelname)s - %(name)s -   %(message)s",
            datefmt = "%m/%d/%Y %H:%M:%S",
            # level = logging.INFO if self.local_rank in [-1, 0] else logging.WARN,
            level = llevel
        )  
        
        ### 匯入 tokenizer,model ###
        hf_cache_dir = str(Path.home()) + "/senti_c/huggingface/"
        if not os.path.exists(hf_cache_dir):
            os.makedirs(hf_cache_dir)
            
        self.tokenizer = BertTokenizer.from_pretrained(tokenizer, cache_dir = hf_cache_dir)
        
        
        if model_name_or_path in SENTENCE_CLASSIFICATION_MODEL:
            ## 下載與解壓縮預設模型並存放於與預設模型同一資料夾 ##
            model_paths = os.path.split(os.path.realpath(__file__))[0] +  SENTENCE_CLASSIFICATION_MODEL[model_name_or_path] + "/"+ model_cache_folder_name  #取得模型資料夾位置絕對路徑
            
            logger.info(f"Checking model at {model_paths}")
            if not os.path.exists(model_paths):
                createok = False
                try:
                    os.makedirs(model_paths)
                    createok = True
                except Exception as e:
                    logger.warning(f"Encountered error when creating folder {model_paths}: {e}")
                
                if createok == False:                    
                    # tmpdir = tempfile.TemporaryDirectory()
                    # print(f"tmpdir = {tmpdir.name}")                    
                    homedir = str(Path.home())
                    # print(f"homedir = {homedir}")  
                    model_paths = homedir + "/senti_c/" + SENTENCE_CLASSIFICATION_MODEL[model_name_or_path] + "/" + model_cache_folder_name
                    logger.info(f"Try to create model folder at temp dir: {model_paths}")
                    os.makedirs(model_paths)
                
            get_toolkit_models(model_paths,"sentence",model_name_or_path)
            self.model = MyNewBertForSequenceClassification.from_pretrained(model_paths)
        else:
            self.model = MyNewBertForSequenceClassification.from_pretrained(model_name_or_path)
        
        
        ### 設定 cpu,gpu ###
        
        if self.local_rank == -1 or no_cuda:
            self.device = torch.device("cuda" if torch.cuda.is_available() and not no_cuda else "cpu")
            self.n_gpu = 0 if no_cuda else torch.cuda.device_count()
        else:
            torch.cuda.set_device(self.local_rank)
            self.device = torch.device("cuda", self.local_rank)
            torch.distributed.init_process_group(backend="nccl")
            self.n_gpu = 1
        
        self.model.to(self.device) 

        


    def _check_input_list(self,input_data_or_path):                                    
        '''檢查輸入資料(list)的格式'''
        
        data = []
        if type(input_data_or_path) != list:
            raise Exception("輸入資料型別錯誤！請重新確認！")
        elif len(input_data_or_path) == 0 :
            raise Exception("輸入資料大小為0！請重新確認！")
        else:
            data = input_data_or_path 
        return data

    
    def _check_input_tsv(self,input_data_or_path):                                    
        '''檢查輸入資料(tsv)的格式；並將tsv資料轉成list'''
        
        data = []
        
        try:
            if not os.path.isfile(input_data_or_path):
                raise Exception        
        except:
            raise Exception("輸入資料不存在或並非檔案格式！請重新確認！")
        
        
        processor = SentenceProcessor()
        data = processor.get_test_examples_from_tsv(input_data_or_path)
        return data 
    

    def _read_input_list(self,input_data,output_dir=None,overwrite_cache=True):                                 
        '''產生符合模型需求的資料格式'''
        
        ## 獲取examples ##
        processor = SentenceProcessor()
        examples = processor.get_test_examples_from_list(input_data)
        
        if output_dir is not None: #表示有指定輸出路徑
            cached_features_file = os.path.join(
                output_dir,
                "cached_test",
            )
        else: #將路徑改成當前目錄
            cached_features_file = os.path.join(
                ".",
                "cached_test",
            )
        
        if os.path.exists(cached_features_file) and not overwrite_cache:
            logger.info("載入先前的 cache 檔案 %s", cached_features_file)
            features = torch.load(cached_features_file)
        else:    
            logger.info("創建 cache 檔案 %s", cached_features_file)
                
            ## 將資料轉成特徵 ##
            labels = processor.get_labels()
            features = convert_examples_to_features(
                examples,
                self.tokenizer,
                label_list = labels,
                max_length = 256,
                pad_token = self.tokenizer.convert_tokens_to_ids([self.tokenizer.pad_token])[0],
                logging_display = self.logging_display,
            )  
            
            
            if self.local_rank in [-1, 0]:                
                logger.info("儲存 cache 檔案 %s", cached_features_file)
                torch.save(features, cached_features_file)
        
        ## 將特徵轉成tensors並建成Dataset ##
        all_input_ids = torch.tensor([f.input_ids for f in features], dtype=torch.long)
        all_attention_mask = torch.tensor([f.attention_mask for f in features], dtype=torch.long)
        all_token_type_ids = torch.tensor([f.token_type_ids for f in features], dtype=torch.long)
        
        dataset = TensorDataset(all_input_ids, all_attention_mask, all_token_type_ids)
        return dataset
            
    
    def _get_aggregate_pred_class(self,preds):
        '''
        將同一輸入文本的各句子的預測類別作整合，獲得單一預測類別
        整合原則 : 輸入文本的各句子中，若有正有負則歸類為衝突；如果全部只有一種類別則就直接歸為此類別；如果有正面/負面加上中性，則歸為正面/負面；如果有任一衝突類別則歸為衝突類別
        '''
        
        if "衝突" in preds: 
            return "衝突"
        elif ("負面" in preds) and ("正面" in preds):
            return "衝突"
        elif ("負面" in preds) and ("中性" in preds): 
            return "負面"
        elif ("正面" in preds) and ("中性" in preds):
            return "正面"
        else:  #剩下的狀況只會是所有的都是同一類別
            return preds[0] 

    
    def predict(self,
                input_data_or_path,
                input_mode = "list",
                run_split = False,
                aggregate_strategy = False,
                batch_size = 8,
                output_mode = "return_result",
                output_dir = None,          
                threshold = 0.5,
                overwrite_cache = True,
               ):
        '''
        預測句子情感的專用函數。(注意初始化class後，可傳入不同參數來重複呼叫此函數進行預測)
        參數包含：
        input_data_or_path : 輸入數據，和input_mode相關聯與對應，默認為輸入list型別  
        input_mode : 輸入模式選擇，分為"list"或"tsv"，默認"list"  => 考慮句子和屬性部分統一成txt
        run_split : 是否對每筆輸入資料執行斷句，默認"False"
        aggregate_strategy : 如果執行斷句，是否將斷句後的各句預測情感匯總(True)、或分別呈現(False)，默認"False"
        batch_size : 一次丟入多筆待預測資料到模型，默認"8"
        output_mode : 輸出模式選擇，分為"寫入檔案+回傳預測變數(write_file)"或"單純回傳預測變數(return_result)"，默認"單純回傳預測變數(return_result)"
        output_dir : 輸出檔案路徑，可選參數
        threshold : 將正面和負面機率轉為類別的臨界值，默認"0.5"
        overwrite_cache : 是否覆寫預測資料的cache檔案，默認為True，若要重複預測同樣的內容，可將此參數改為False，程式會讀取先前儲存的cache檔案
        '''

        
        all_vals = locals()
        del all_vals["self"]  #不須顯示此項
        logger.info("predict函數的所有參數： %s",all_vals)
        logger.info("開始讀取與檢查輸入數據！")
        
        if not input_data_or_path: #如果沒有任何值
            raise Exception("缺少輸入資料！請重新確認！")
        
        if input_mode == 'list':
            ### 檢查輸入資料格式 ###
            input_data = self._check_input_list(input_data_or_path)
        else:
            ### 檢查輸入資料格式，並獲得list格式的資料 ###
            input_data = self._check_input_tsv(input_data_or_path)    
            
        
        ## 依序處理每筆資料 ##
        all_data = [] #儲存所有要丟入模型的資料
        input_data_chg = []  #儲存每個句子對應的原始文本內容，這是給後面輸出時所用
        split_index = [] #儲存原本輸入文本所對應的累積句子數 (因為當輸入文本經過斷句後，可能一個輸入文本包含多個句子，為了方便後續作匯總情感，所以儲存這項) => e.g. 輸入文本有三個，斷句後分別有1,3,2個句子，則對應的 split_index 為[1,4,6]，到時要辨別的話，便是用0:1,1:4,4:6來取得
        current_counts = 0 #目前累積句子數
        
        for data in input_data: 
            ## 執行斷句 ##
            if run_split:
                sentences = split_text_from_input_data(data)  
            else:
                sentences = [data]
            
            sentences = [line for line in sentences if (len(line) > 0 and not line.isspace())] #去除整個句子中為空白的句子 
            
            all_data.extend(sentences)
            current_counts += len(sentences)
            split_index.append(current_counts)

            input_data_chg.extend([data] * len(sentences))
        
        try:
            assert len(all_data) == current_counts == len(input_data_chg)
        except:
            raise Exception("處理過程中發生錯誤！請聯繫開發者！")
        
        
        
        logger.info("原本的總輸入文本數目：%s",len(input_data))
        logger.info("處理後的總文本/句子數目：%s",current_counts)
        
        
        ## 將資料轉成模型需要的dataset ##  
        logger.info("開始處理輸入資料為模型需求格式！")
        test_dataset = self._read_input_list(all_data,output_dir,overwrite_cache)
         
            
        ## 實際跑模型獲得預測結果 ##
        test_batch_size = batch_size * max(1, self.n_gpu)
        test_sampler = SequentialSampler(test_dataset) if self.local_rank == -1 else DistributedSampler(test_dataset)
        test_dataloader = DataLoader(test_dataset, sampler=test_sampler, batch_size=test_batch_size)
        
        if self.n_gpu > 1 and not isinstance(self.model, torch.nn.DataParallel):
            self.model = torch.nn.DataParallel(self.model)
        
        
        logger.info("***** 開始進行預測  *****")
        logger.info("  資料數量 = %d", len(test_dataset))
        logger.info("  Batch大小 = %d", test_batch_size)
        
        preds = None  #最後接起來維度會是:(總樣本數, 2)
        for batch in tqdm(test_dataloader, desc="Predicting"):
            self.model.eval()
            batch = tuple(t.to(self.device) for t in batch)
            
            with torch.no_grad():
                inputs = {"input_ids": batch[0], "attention_mask": batch[1] ,"token_type_ids": batch[2]} #因為要預測，所以沒有labels

                outputs = self.model(**inputs)
                logits = outputs[0] #因為沒有label

            if preds is None:
                preds = logits.detach().cpu().numpy()  #logits維度:(bsz,num-lables)
            else:
                preds = np.append(preds, logits.detach().cpu().numpy(), axis = 0) 
        
        
        logger.info("已完成預測！開始整理預測結果！")
                
        ## 整理預測值 ##
        preds = sigmoid_array(preds)  #經過sigmoid 
        new_preds = np.zeros_like(preds)  #生成相同形狀和資料類型的array(這是要放轉為標籤後的,預設放0)
        new_preds [preds > threshold] = 1  #將大於threshold的標籤改為1
        
        
        ## 轉成類別 ##
        toolkit_preds = []
        for i in range(len(new_preds)):
            pos = int(new_preds[i][0])
            neg = int(new_preds[i][1])

            if pos == 1 and neg == 0:
                toolkit_preds.append("正面")
            elif pos == 0 and neg == 1:
                toolkit_preds.append("負面")    
            elif pos == 0 and neg == 0:
                toolkit_preds.append("中性")
            else:  
                toolkit_preds.append("衝突")
        
        ## 整理要輸出或返回的預測結果 ##
        if run_split:
            ## 彙總情感 ##
            if aggregate_strategy: 
                split_index.insert(0,0) #最前面增加0、方便後續處理

                all_text_preds = []  #儲存每個輸入文本對應的預測類別
                for i in range(1,len(split_index)):
                    # 取得原始的一筆輸入文本的各句子情感 #
                    text_preds = toolkit_preds[split_index[i-1]:split_index[i]] 
                    result = self._get_aggregate_pred_class(text_preds) #得到綜合後的預測類別
                    all_text_preds.append(result)

                #第一欄為原始文本，第二欄為綜合後的預測結果
                pred_result = pd.DataFrame({"Inputs":input_data ,"Preds":all_text_preds})
            else:
                #第一欄為原始文本，第二欄為斷句後句子，第三欄為各句子預測結果
                pred_result = pd.DataFrame({"Inputs":input_data_chg ,"Sentences":all_data,  "Preds":toolkit_preds})
                
        else:
            #第一欄為原始文本，第二欄為對應預測結果
            pred_result = pd.DataFrame({"Inputs":input_data ,"Preds":toolkit_preds})
            
        
        ## 返回/寫入預測結果 ##    
        if output_mode == 'write_file':
            if output_dir is not None:
                output_test_predictions_file = os.path.join(output_dir,"predictions.tsv")
            else:
                output_test_predictions_file = "predictions.tsv"  #表示會放在與目前位置同一路徑
                
            
            logger.info("將預測結果寫入檔案 %s",output_test_predictions_file)
                    
            pred_result.to_csv(output_test_predictions_file,sep='\t',index=False,header=True)
        
        return pred_result  #不管哪種模式都會回傳預測結果的變數，使用者可以依據需求選擇是否使用 
        
        
        
            
            
        
          



from senti_c import AspectSentimentAnalysis

aspect_classifier = AspectSentimentAnalysis()

test_data = ["我很喜歡這家店！超級無敵棒！","這個服務生很不親切..."]   
result = aspect_classifier.predict(test_data, output_result="all")

print("\n *** \n")
print("Aspect and Sentiment", result['AspectTermAndSentimentExtraction'])

nseg = len(result['InputWords'])
# result['AspectTermTags']
for seg in range(nseg):
    print(f"   Sentence {seg}")
    a1 = result['InputWords'][seg]
    a2 = result['AspectTermAndSentimentTags'][seg]
    for x1, x2 in zip(a1, a2):
        print(x1, x2)


# coding=utf-8
# Some of structures or codes in this script are referenced from HuggingFace Inc. team. 

# Copyright 2018 The Google AI Language Team Authors and The HuggingFace Inc. team.
# Copyright (c) 2018, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
'''通用的(句子/屬性)一些函數'''

import random
import numpy as np
import torch
import zipfile
import requests
import os


def _download_file_from_google_drive(id, destination):
    URL = "https://docs.google.com/uc?export=download"

    session = requests.Session()

    response = session.get(URL, params = { 'id' : id }, stream = True)
    token = _get_confirm_token(response)

    if token:
        params = { 'id' : id, 'confirm' : token }
        response = session.get(URL, params = params, stream = True)

    _save_response_content(response, destination)    

def _get_confirm_token(response):
    for key, value in response.cookies.items():
        if key.startswith('download_warning'):
            return value

    return None

def _save_response_content(response, destination):
    CHUNK_SIZE = 32768

    with open(destination, "wb") as f:
        for chunk in response.iter_content(CHUNK_SIZE):
            if chunk: # filter out keep-alive new chunks
                f.write(chunk)

def get_toolkit_models(model_path,types="sentence",varients="default"):
    '''用來下載與解壓縮模型'''
    
    #先下載壓縮檔 :
    #壓縮檔包含模型相關資料
    ## 程式來自 https://stackoverflow.com/questions/25010369/wget-curl-large-file-from-google-drive/39225039#39225039
    if types == "sentence":
        if varients == "default":
            file_id = '1DnSE7duvvJsqTTLWelGWGW9E8EH47ODp' 
        else:
            file_id = '1mwWLrjiYb7YyeHd5EtWKrHz_I0_YpYF-' 
    else:
        if varients == "default":
            file_id = '1HHjpR9Am9a2JgDRrqVWh6t5BRRjlqk2u' 
        else:
            file_id = '1a2RtKsGtgK6c3yCE_ABpZZZ18VcvZApv'
            
    destination = model_path+'/models.zip'
    _download_file_from_google_drive(file_id, destination)
    
    #解壓縮
    with zipfile.ZipFile(destination,'r') as zips:
        for fl_name in zips.namelist():
            zips.extract(fl_name, model_path)    
        
    #刪除壓縮檔    
    try:
        os.remove(destination)
    except:
        raise Exception("壓縮檔 %s 不存在！請重新確認！",destination)
        

def get_domain_embedding(file_path):
    '''用來下載領域詞向量檔案'''
    
    ## 程式來自 https://stackoverflow.com/questions/25010369/wget-curl-large-file-from-google-drive/39225039#39225039
    file_id = '1wVJUNhbHraehDyzo_2s5apWEoYmzRuUS'
            
    if not os.path.isfile(file_path):
        _download_file_from_google_drive(file_id, file_path)

        
def split_text_from_input_data(text):
    '''斷句程式'''   
    
    start = 0
    pos = 0
    length = len(text)  #文本長度
    sentences = []
    inQuote = False

    while (pos < length):  #一個個遍歷
        c = text[pos]

        if (c == '。' or c == '？' or c == '！' or c =='!' or c== '?' or c=='~' or c =='～'):

            if (pos >= length-1):  # 以句號、問號和感嘆號結束文本，成句
                sentence = text[start:pos+1]
                sentences.append(sentence);

                pos = pos+1
                start = pos

            else:  #句號、問號和感嘆號還有内容

                nextC = text[pos+1]

                if (nextC == '”' or nextC == '〞' or nextC =='′' or nextC == '」'): #句號、問號和感嘆號在引號内，和引號一起成句
                    sentence = text[start: pos+2]
                    sentences.append(sentence)

                    pos += 2
                    start = pos

                    inQuote = False

                elif (nextC == '。' or nextC == '？' or nextC == '！' or nextC =='!' or nextC== '?'or nextC=='~' or nextC =='～'): #多個重複的。？！
                    pos = pos + 1 

                else:
                    if (~inQuote):   #句號、問號和感嘆號不在引號内，成句
                        sentence = text[start:pos + 1]
                        sentences.append(sentence)

                        pos = pos+1
                        start = pos
                    else:  #找引號
                        pos = pos + 1

        else:
            if (c == '“' or c == '「'):
                inQuote = True
            elif (c == '”' or c == '〞' or c =='′' or c == '」'):
                inQuote = False
            pos = pos + 1


    if (pos > start):    # 還有剩餘内容，成句 (如文本沒有什麼句號之類做結尾)
        sentence = text[start:pos]
        sentences.append(sentence)

    return sentences    







# -*- coding: utf-8 -*-
from .sentence_sentiment_analysis import SentenceSentimentClassification
from .sentence_model_finetuning import SentenceSentimentModel

from .aspect_sentiment_analysis import AspectSentimentAnalysis
from .aspect_model_finetuning import AspectSentimentModel



# coding=utf-8
# Some of structures or codes in this script are referenced from HuggingFace Inc. team. 

# Copyright 2018 The Google AI Language Team Authors and The HuggingFace Inc. team.
# Copyright (c) 2018, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
'''屬性情感分類模型的微調/訓練'''

import glob
import json
import logging
import os,sys
import random
import copy
import numpy as np
import torch
import torch.nn as nn 
from torch.utils.data import DataLoader, RandomSampler, SequentialSampler, TensorDataset
from torch.utils.data.distributed import DistributedSampler
from tqdm import tqdm, trange


from transformers import (
    WEIGHTS_NAME,
    AdamW,
    BertTokenizer,  
    BertConfig,
    BertModel,
    BertPreTrainedModel,
    get_linear_schedule_with_warmup,
)

try:
    from torch.utils.tensorboard import SummaryWriter
except ImportError:
    from tensorboardX import SummaryWriter

from .model_structure import MyNewBertForTokenClassification  
from .utils import AspectProcessor, aspect_convert_examples_to_features as convert_examples_to_features, evaluate_ote, evaluate_ts, ASPECT_LABEL_MAP, SENTIMENT_LABEL_MAP, chg_labels_to_aspect_and_sentiment


logger = logging.getLogger(__name__)  

DEFAULT_ASPECT_CLASSIFICATION_PRETRAIN_MODEL = 'bert-base-chinese'  


def _set_seed(seed,n_gpu):
    '''設定seed'''
    
    random.seed(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)
    if n_gpu > 0:
        torch.cuda.manual_seed_all(seed)

        
class AspectSentimentModel:
    '''屬性情感分類模型的訓練：使用設定的參數來訓練/微調預測模型'''  
    
    def __init__(self,
                 model_name_or_path=DEFAULT_ASPECT_CLASSIFICATION_PRETRAIN_MODEL,
                 config_name="",
                 tokenizer_name="",
                 cache_dir="",
                 no_cuda=False,
                 local_rank=-1,
                 logging_display=True,
                 seed=42,
                ):  
        '''
        設定屬性情感分類模型訓練相關參數：
        model_name_or_path : 預訓練的模型，可以為路徑或是huggingface團隊支援的模型名稱，默認為"bert-base-chinese"
        config_name : 如果不想使用預訓練模型的config，可以自己撰寫新的config檔案，這裡需放名稱或路徑，可選參數
        tokenizer_name : 如果不想使用預訓練模型的tokenizer，可指定別的名稱或路徑，可選參數
        cache_dir : 放置下載的預訓練模型的位置，可選參數
        no_cuda : 是否避免使用gpu，默認"False"
        local_rank : 是否使用平行化運算，默認"-1"
        logging_display : 設置是否顯示logging，默認為"True"
        seed : 隨機種子，默認為"42"
        '''
        
        ### 設定 cpu,gpu ###
        self.local_rank = local_rank
        
        if self.local_rank == -1 or no_cuda:
            self.device = torch.device("cuda" if torch.cuda.is_available() and not no_cuda else "cpu")
            self.n_gpu = 0 if no_cuda else torch.cuda.device_count()
        else:
            torch.cuda.set_device(self.local_rank)
            self.device = torch.device("cuda", self.local_rank)
            torch.distributed.init_process_group(backend="nccl")
            self.n_gpu = 1

        ### 設定 logging ###
        self.logging_display = logging_display
        
        logging.basicConfig(
            format="%(asctime)s - %(levelname)s - %(name)s -   %(message)s",
            datefmt="%m/%d/%Y %H:%M:%S",
            level=logging.INFO if self.local_rank in [-1, 0] else logging.WARN,
        )   
        
        
        if self.logging_display:
            logger.warning(
                "Process rank: %s, device: %s, n_gpu: %s, distributed training: %s",
                self.local_rank,
                self.device,
                self.n_gpu,
                bool(self.local_rank != -1),
            )

        
        ### 設定 seed ###
        self.seed = seed
        _set_seed(self.seed,self.n_gpu)
        
        
        ### 匯入 config,tokenizer,model ###
        self.model_name_or_path = model_name_or_path
        
        if self.local_rank not in [-1, 0]:
            torch.distributed.barrier() 
        
        self.config = BertConfig.from_pretrained(
            config_name if config_name else model_name_or_path,
            num_labels=7,
            cache_dir=cache_dir if cache_dir else None,
        ) 
        self.tokenizer = BertTokenizer.from_pretrained(
            tokenizer_name if tokenizer_name else model_name_or_path,
            cache_dir=cache_dir if cache_dir else None,
        )
        self.model = MyNewBertForTokenClassification.from_pretrained(
            model_name_or_path,
            from_tf=bool(".ckpt" in model_name_or_path),
            config=self.config,
            cache_dir=cache_dir if cache_dir else None,
        )

        if self.local_rank == 0:
            torch.distributed.barrier() 

        
    def _train(self,saved_args,train_dataset):  
        '''訓練模型的主要函數'''  
        
        ## 取出模型部分 ##
        model = saved_args["local_model"]  
        
        if self.local_rank in [-1, 0]:   
            tb_writer = SummaryWriter()   

        saved_args['train_batch_size'] = saved_args["per_gpu_train_batch_size"] * max(1, self.n_gpu)
        train_sampler = RandomSampler(train_dataset) if self.local_rank == -1 else DistributedSampler(train_dataset)
        train_dataloader = DataLoader(train_dataset, sampler=train_sampler, batch_size=saved_args['train_batch_size'])

        if saved_args["max_steps"] > 0:
            t_total = saved_args["max_steps"]
            saved_args["num_train_epochs"] = saved_args["max_steps"] // (len(train_dataloader) // saved_args["gradient_accumulation_steps"]) + 1
        else:
            t_total = len(train_dataloader) // saved_args["gradient_accumulation_steps"] * saved_args["num_train_epochs"]

        
        no_decay = ["bias", "LayerNorm.weight"]
        optimizer_grouped_parameters = [
            {
                "params": [p for n, p in model.named_parameters() if not any(nd in n for nd in no_decay)],
                "weight_decay": saved_args["weight_decay"],
            },
            {"params": [p for n, p in model.named_parameters() if any(nd in n for nd in no_decay)], "weight_decay": 0.0},
        ]

        optimizer = AdamW(optimizer_grouped_parameters, lr=saved_args["learning_rate"], eps=saved_args["adam_epsilon"])
        scheduler = get_linear_schedule_with_warmup(
            optimizer, num_warmup_steps=saved_args["warmup_steps"], num_training_steps=t_total
        )

        
        if os.path.isfile(os.path.join(self.model_name_or_path, "optimizer.pt")) and os.path.isfile(
            os.path.join(self.model_name_or_path, "scheduler.pt")
        ):
            optimizer.load_state_dict(torch.load(os.path.join(self.model_name_or_path, "optimizer.pt")))
            scheduler.load_state_dict(torch.load(os.path.join(self.model_name_or_path, "scheduler.pt")))

        if self.n_gpu > 1:
            model = torch.nn.DataParallel(model)

        ## 平行化處理 ##
        if self.local_rank != -1:
            model = torch.nn.parallel.DistributedDataParallel(
                model, device_ids=[self.local_rank], output_device=self.local_rank, find_unused_parameters=True,
            )

        ## 以下開始訓練 ##
        if self.logging_display:
            logger.info("***** Running training *****")
            logger.info("  Num examples = %d", len(train_dataset))
            logger.info("  Num Epochs = %d", saved_args["num_train_epochs"])
            logger.info("  Instantaneous batch size per GPU = %d", saved_args["per_gpu_train_batch_size"])
            logger.info(
                "  Total train batch size (w. parallel, distributed & accumulation) = %d",
                saved_args['train_batch_size']
                * saved_args["gradient_accumulation_steps"]
                * (torch.distributed.get_world_size() if self.local_rank != -1 else 1),
            )
            logger.info("  Gradient Accumulation steps = %d", saved_args["gradient_accumulation_steps"])
            logger.info("  Total optimization steps = %d", t_total)

        global_step = 0
        epochs_trained = 0
        steps_trained_in_current_epoch = 0
        
        ## 檢查是否從某個檢查點繼續訓練 ##
        if os.path.exists(self.model_name_or_path):
            try:
                global_step = int(self.model_name_or_path.split("-")[-1].split("/")[0])
            except ValueError:
                global_step = 0
            epochs_trained = global_step // (len(train_dataloader) // saved_args["gradient_accumulation_steps"])
            steps_trained_in_current_epoch = global_step % (len(train_dataloader) // saved_args["gradient_accumulation_steps"])
            
            if self.logging_display:
                logger.info("  Continuing training from checkpoint, will skip to saved global_step")
                logger.info("  Continuing training from epoch %d", epochs_trained)
                logger.info("  Continuing training from global step %d", global_step)
                logger.info("  Will skip the first %d steps in the first epoch", steps_trained_in_current_epoch)

        tr_loss, logging_loss = 0.0, 0.0

        model.zero_grad()
        train_iterator = trange(
            epochs_trained, int(saved_args["num_train_epochs"]), desc="Epoch", disable=self.local_rank not in [-1, 0],
        )
        
        _set_seed(self.seed,self.n_gpu)  ## 放這以便重製結果
        for _ in train_iterator:  # 保留epoch的tqdm
            epoch_iterator = tqdm(train_dataloader, desc="Iteration", disable=self.local_rank not in [-1, 0])
            for step, batch in enumerate(epoch_iterator):
                if steps_trained_in_current_epoch > 0:
                    steps_trained_in_current_epoch -= 1
                    continue

                model.train()
                batch = tuple(t.to(self.device) for t in batch)
                    
                inputs = {"input_ids": batch[0], "attention_mask": batch[1], "token_type_ids" : batch[2], "labels": batch[3]}
                
                outputs = model(**inputs)
                loss = outputs[0]  

                if self.n_gpu > 1:
                    loss = loss.mean()  # 因為平行化處理
                if saved_args["gradient_accumulation_steps"] > 1:
                    loss = loss / saved_args["gradient_accumulation_steps"]

                loss.backward()

                tr_loss += loss.item()
                if (step + 1) % saved_args["gradient_accumulation_steps"] == 0:
                    torch.nn.utils.clip_grad_norm_(model.parameters(), saved_args["max_grad_norm"])

                    optimizer.step()
                    scheduler.step()  
                    model.zero_grad()
                    global_step += 1


                    if self.local_rank in [-1, 0] and saved_args["logging_steps"] > 0 and global_step % saved_args["logging_steps"] == 0 :
                        logs = {}

                        if (self.local_rank == -1 and saved_args["evaluate_during_training"]): 
                            results = self._evaluate(model,saved_args)
                            
                            for key, value in results.items():
                                eval_key = "eval_{}".format(key)
                                logs[eval_key] = value
                            
  
                        loss_scalar = (tr_loss - logging_loss) / saved_args["logging_steps"]
                        learning_rate_scalar = scheduler.get_lr()[0]
                        logs["learning_rate"] = learning_rate_scalar
                        logs["train_eval_loss"] = loss_scalar   
                        logging_loss = tr_loss
                    
                        for key, value in logs.items():
                            tb_writer.add_scalar(key, value, global_step)
                        if self.logging_display:
                            print(json.dumps({**logs, **{"step": global_step}}))

                        
                    if self.local_rank in [-1, 0] and saved_args["save_steps"] > 0 and global_step % saved_args["save_steps"] == 0:
                        ## 儲存模型 checkpoint ##
                        checkpoint_output_dir = os.path.join(saved_args["output_path"], "checkpoint-{}".format(global_step))
                        if not os.path.exists(checkpoint_output_dir):
                            os.makedirs(checkpoint_output_dir)
                        model_to_save = (
                            model.module if hasattr(model, "module") else model
                        )  
                        model_to_save.save_pretrained(checkpoint_output_dir)
                        self.tokenizer.save_pretrained(checkpoint_output_dir)

                        torch.save(saved_args, os.path.join(checkpoint_output_dir, "training_args.bin"))
                        if self.logging_display:
                            logger.info("Saving model checkpoint to %s", checkpoint_output_dir)

                        torch.save(optimizer.state_dict(), os.path.join(checkpoint_output_dir, "optimizer.pt"))
                        torch.save(scheduler.state_dict(), os.path.join(checkpoint_output_dir, "scheduler.pt"))
                        if self.logging_display:
                            logger.info("Saving optimizer and scheduler states to %s", checkpoint_output_dir)

                if saved_args["max_steps"] > 0 and global_step > saved_args["max_steps"]:
                    epoch_iterator.close()
                    break
            if saved_args["max_steps"] > 0 and global_step > saved_args["max_steps"]:
                train_iterator.close()
                break
                
        if self.local_rank in [-1, 0]:
            tb_writer.close()

        return global_step, tr_loss / global_step , model , saved_args

    
    def _evaluate(self, model, saved_args, prefix=""):
        '''對模型進行驗證'''
 
        eval_dataset, all_evaluate_label_ids = self._load_and_cache_examples(saved_args, evaluate=True) 

        if not os.path.exists(saved_args["output_dir"]) and self.local_rank in [-1, 0]:
            os.makedirs(saved_args["output_dir"])

        saved_args["eval_batch_size"] = saved_args["per_gpu_eval_batch_size"] * max(1, self.n_gpu)
        eval_sampler = SequentialSampler(eval_dataset) 
        eval_dataloader = DataLoader(eval_dataset, sampler=eval_sampler, batch_size=saved_args["eval_batch_size"])

        # 多個GPU時
        if self.n_gpu > 1 and not isinstance(model, torch.nn.DataParallel):
            model = torch.nn.DataParallel(model)

        ## 以下開始對模型進行驗證 ##
        if self.logging_display:
            logger.info("***** Running evaluation {} *****".format(prefix))
            logger.info("  Num examples = %d", len(eval_dataset))
            logger.info("  Batch size = %d", saved_args["eval_batch_size"])

        eval_loss = 0.0
        aspect_eval_loss = 0.0
        sentiment_eval_loss = 0.0
        nb_eval_steps = 0
        
        aspect_preds = None
        sentiment_preds = None
        aspect_out_label_ids = None
        sentiment_out_label_ids = None

        for batch in tqdm(eval_dataloader, desc="Evaluating"):
            model.eval()
            batch = tuple(t.to(self.device) for t in batch)

            with torch.no_grad():
                inputs = {"input_ids": batch[0], "attention_mask": batch[1], "token_type_ids" : batch[2], "labels": batch[3]}
                outputs = model(**inputs)
                tmp_eval_loss, tmp_aspect_loss,tmp_sentiment_loss, aspect_logits, sentiment_logits = outputs[:5]

                if self.n_gpu > 1:
                    tmp_eval_loss = tmp_eval_loss.mean()  
                    tmp_aspect_loss = tmp_aspect_loss.mean()
                    tmp_sentiment_loss = tmp_sentiment_loss.mean()
                
            nb_eval_steps += 1
            
            if aspect_preds is None:
                aspect_preds = aspect_logits.detach().cpu().numpy()
                sentiment_preds = sentiment_logits.detach().cpu().numpy()
                
                aspect_labels, sentiment_labels = chg_labels_to_aspect_and_sentiment(inputs["labels"])
                aspect_out_label_ids , sentiment_out_label_ids = aspect_labels.detach().cpu().numpy(), sentiment_labels.detach().cpu().numpy()
                
            else:
                aspect_preds = np.append(aspect_preds, aspect_logits.detach().cpu().numpy(), axis=0)
                sentiment_preds = np.append(sentiment_preds, sentiment_logits.detach().cpu().numpy(), axis=0)
                
                aspect_labels,sentiment_labels = chg_labels_to_aspect_and_sentiment(inputs["labels"])
                
                aspect_out_label_ids = np.append(aspect_out_label_ids, aspect_labels.detach().cpu().numpy(), axis=0)
                sentiment_out_label_ids = np.append(sentiment_out_label_ids, sentiment_labels.detach().cpu().numpy(), axis=0)

                
        eval_loss = eval_loss / nb_eval_steps
        aspect_eval_loss = aspect_eval_loss / nb_eval_steps
        sentiment_eval_loss = sentiment_eval_loss / nb_eval_steps

        aspect_preds = np.argmax(aspect_preds, axis=2)
        sentiment_preds = np.argmax(sentiment_preds, axis=2)
        
        ## 以下計算評估標準 ##
        aspect_out_label_list = [[] for _ in range(aspect_out_label_ids.shape[0])]
        aspect_preds_list = [[] for _ in range(aspect_out_label_ids.shape[0])]

        # 以下兩者是為了去計算整體ABSA評估標準所用
        absa_out_label_list = [[] for _ in range(sentiment_out_label_ids.shape[0])]
        absa_preds_list = [[] for _ in range(sentiment_out_label_ids.shape[0])]
        
        for i in range(aspect_out_label_ids.shape[0]):
            evaluate_label_ids = all_evaluate_label_ids[i] #取得該句子中有效的位置索引
            for j in range(aspect_out_label_ids.shape[1]):
                if j in evaluate_label_ids: 
                    asp_true = ASPECT_LABEL_MAP[aspect_out_label_ids[i][j]]
                    asp_pred = ASPECT_LABEL_MAP[aspect_preds[i][j]]
                    sent_true = SENTIMENT_LABEL_MAP[sentiment_out_label_ids[i][j]]
                    sent_pred = SENTIMENT_LABEL_MAP[sentiment_preds[i][j]]

                    aspect_out_label_list[i].append(asp_true)
                    aspect_preds_list[i].append(asp_pred)

                    # 將屬性標記和情感標記標籤結合
                    absa_out_label_list[i].append(asp_true+"-"+sent_true)
                    absa_preds_list[i].append(asp_pred+"-"+sent_pred)
        
        ote_scores = evaluate_ote(gold_ot=aspect_out_label_list, pred_ot=aspect_preds_list)
        ts_scores = evaluate_ts(gold_ts=absa_out_label_list, pred_ts=absa_preds_list)

        results = {
            "loss": eval_loss,
            "aspect_loss": aspect_eval_loss,
            "sentiment_loss": sentiment_eval_loss,
        }
        results.update(ote_scores)
        results.update(ts_scores)

 
        if self.logging_display:
            logger.info("***** Eval results {} *****".format(prefix))
            for key in sorted(results.keys()):
                logger.info("  %s = %s", key, str(results[key]))

        return results

    
    def _load_and_cache_examples(self, saved_args, evaluate=False): 
        '''讀取資料集&轉成特徵'''
        
        if self.local_rank not in [-1, 0] and not evaluate:
            torch.distributed.barrier()  

        processor = AspectProcessor()
        
        mode = 'dev' if evaluate else 'train'
        cached_features_file = os.path.join(
            saved_args["data_dir"],
            "cached_{}_{}_{}".format(
                mode,
                list(filter(None, self.model_name_or_path.split("/"))).pop(),
                str(saved_args["max_seq_length"]),
            ),
        )
        
        if os.path.exists(cached_features_file) and not saved_args["overwrite_cache"]:
            if self.logging_display:
                logger.info("載入先前的 cache 檔案 %s", cached_features_file)
            features = torch.load(cached_features_file)
        else:
            if self.logging_display:
                logger.info("創建 cache 檔案 %s", cached_features_file)
            
            ## 獲得 examples ##
            examples = processor.get_train_dev_examples(saved_args["data_dir"],mode) 
            
            labels=processor.get_labels()
            ## 將資料轉成特徵 ##
            features = convert_examples_to_features(
                examples,
                self.tokenizer,
                label_list=labels,
                max_length=saved_args["max_seq_length"],
                pad_token=self.tokenizer.convert_tokens_to_ids([self.tokenizer.pad_token])[0],
                logging_display=self.logging_display,
            )
            if self.local_rank in [-1, 0]:
                if self.logging_display:
                    logger.info("儲存 cache 檔案 %s", cached_features_file)
                torch.save(features, cached_features_file)
               
        if self.local_rank == 0 and not evaluate:
            torch.distributed.barrier()  

        ## 將特徵轉成tensors並建成Dataset ##
        all_input_ids = torch.tensor([f.input_ids for f in features], dtype=torch.long)
        all_attention_mask = torch.tensor([f.attention_mask for f in features], dtype=torch.long)
        all_token_type_ids = torch.tensor([f.token_type_ids for f in features], dtype=torch.long)
        all_labels = torch.tensor([f.label for f in features], dtype=torch.long)
        
        all_evaluate_label_ids = [f.evaluate_label_ids for f in features]  #驗證或預測時會用到

        dataset = TensorDataset(all_input_ids, all_attention_mask, all_token_type_ids, all_labels)
        return dataset, all_evaluate_label_ids

    
    def train(self,
              data_dir,
              output_dir,
              max_seq_length=128,
              evaluate_during_training=False,
              do_eval=True,
              eval_all_checkpoints=False,
              save_best_model=True,
              best_model_metric='absa_micro_f1',
              per_gpu_train_batch_size=8,
              per_gpu_eval_batch_size=8,
              gradient_accumulation_steps=1,
              learning_rate=5e-5,
              weight_decay=0.0,
              adam_epsilon=1e-8,
              max_grad_norm=1.0,
              num_train_epochs=5.0,
              max_steps=-1,
              warmup_steps=0,
              logging_display=True,
              logging_steps=500,
              save_steps=500,
              overwrite_output_dir=False,
              overwrite_cache=False,
             ):
        '''
        使用者可直接使用的函數，用來執行訓練模型的程式。(注意初始化class後，可傳入不同參數來重複呼叫此函數進行訓練，亦即使用同樣的預訓練模型，但用不同資料或訓練參數來進行微調！)
        參數包含：
        data_dir : 放置訓練檔案與驗證檔案的資料夾路徑，注意訓練與驗證檔案須為txt格式，且檔案中內容要符合格式規定(一列包含字、屬性-情感標記，中間用空白隔開；不同資料以空行隔開)
        output_dir : 放置輸出模型與相關檔案的輸出路徑
        max_seq_length : 斷字後最大的長度，默認"128"
        evaluate_during_training : 是否在訓練過程中進行驗證，默認"False"
        do_eval : 是否在訓練結束後對模型進行驗證，默認"True"
        eval_all_checkpoints : 訓練完畢後，是否對每個檢查點進行驗證，默認為"False"
        save_best_model : 對模型/檢查點進行驗證後，是否保存驗證結果最佳的模型至output_dir中的"best_model"資料夾，默認為"True"
        best_model_metric : 判斷驗證時最佳模型的評估標準，默認為"absa_micro_f1"，也可指定其他參數："aspect_precision"/ "aspect_recall" / "aspect_f1" / "absa_macro_neg" / "absa_macro_neu" / "absa_macro_pos" / "absa_macro_f1"/ "absa_micro_precision" / "absa_micro_recall"
        per_gpu_train_batch_size : 訓練時的batch size，默認為"8"
        per_gpu_eval_batch_size : 驗證時的batch size，默認為"8"
        gradient_accumulation_steps : 梯度累積，默認為"1"
        learning_rate : Adam優化器的初始學習率，默認為"5e-5"
        weight_decay : 優化器使用的weight_decay，默認為"0.0"
        adam_epsilon : Adam優化器的epsilon值，默認為"1e-8"
        max_grad_norm : 最大梯度的norm值，默認為"1.0"
        num_train_epochs : 訓練過程中的epoch數目，默認為5.0
        max_steps : 如果設為大於0數值則會覆蓋num_train_epochs，默認為"-1"
        warmup_steps : warmup的steps數目，默認為"0"
        logging_display : 設置是否顯示logging，默認為"True"
        logging_steps : 訓練過程中顯示相關logs的steps間隔數，默認為"500"
        save_steps : 訓練過程中儲存檢查點的steps間隔數，默認為"500"
        overwrite_output_dir : 是否覆寫輸出路徑中的資料，默認為"False"
        overwrite_cache : 是否覆寫放置訓練檔案與驗證檔案的資料夾路徑中的cache資料，默認為"False"
        '''
        
        if self.logging_display:
            all_vals = locals()
            del all_vals["self"]  #不須顯示此項
            logger.info("train函數的所有參數：%s",all_vals)
        
        if (
            os.path.exists(output_dir)
            and os.listdir(output_dir)
            and not overwrite_output_dir  
        ):
            raise Exception("輸出路徑 %s 中已存在其他檔案！需設置overwrite_output_dir！",output_dir)
           
        
        local_model = copy.deepcopy(self.model) #為了讓使用者可以多次呼叫同樣instance的train函數進行不同的模型訓練，將模型進行深層複製、避免訓練過程中相互改動
        local_model.to(self.device)
        
        saved_args = locals() #取得目前變數(包含輸入的內容和前面定義的變數)，是一個dict
        
        ## 獲取訓練資料dataset ##
        train_dataset, _ = self._load_and_cache_examples(saved_args,evaluate=False)
        
        ## 實際執行訓練 ##
        global_step, tr_loss, model, saved_args = self._train(saved_args,train_dataset)
        if self.logging_display:
            logger.info(" global_step = %s, average loss = %s", global_step, tr_loss)
        
        
        ## 將最終訓練好的模型保存至輸出路徑 ##
        if self.local_rank == -1 or torch.distributed.get_rank() == 0:
            if not os.path.exists(output_dir) and self.local_rank in [-1, 0]:
                os.makedirs(output_dir)
                
            if self.logging_display:
                logger.info("保存最終訓練好的模型到 %s", output_dir)
                
            model_to_save = (
                model.module if hasattr(model, "module") else model
            ) 
            model_to_save.save_pretrained(output_dir)   
            self.tokenizer.save_pretrained(output_dir)

            torch.save(saved_args, os.path.join(output_dir, "training_args.bin"))
        
        
        ## 對模型進行驗證 ##
        results = {}
        if do_eval and self.local_rank in [-1, 0]:
            best_name = ""  
            best_result = 0    
            
            checkpoints = [output_dir]
            if eval_all_checkpoints:
                checkpoints = list(
                    os.path.dirname(c) for c in sorted(glob.glob(output_dir + "/**/" + WEIGHTS_NAME, recursive=True))
                )
            
            if self.logging_display:
                logger.info("開始驗證以下檢查點：%s", checkpoints)
            
            for checkpoint in checkpoints:
                global_step = checkpoint.split("-")[-1] if len(checkpoints) > 1 else "final"  
                
                model = MyNewBertForTokenClassification.from_pretrained(checkpoint)
                model.to(self.device)
                result = self._evaluate(model,saved_args,prefix=global_step)  
                
                if result[best_model_metric] >= best_result:  
                    best_result = result[best_model_metric]
                    best_name = checkpoint
                
                result = {"{}_{}".format(global_step, k): v for k, v in result.items()}
                results.update(result)
                
            ## 將驗證結果寫入檔案 ##
            output_eval_file = os.path.join(output_dir, "eval_results.txt")  
            with open(output_eval_file, "w") as writer:   
                for key in sorted(results.keys()):
                    writer.write("{} = {}\n".format(key, str(results[key])))
            
            
            ## 將最佳的模型/檢查點保存至輸出路徑中的 "best_model" 資料夾下，使用者可於後續至指定位置取得該模型來進行預測 ##
            if save_best_model:
                if self.logging_display:
                    logger.info("保存最佳模型： %s ", best_name)
                    
                model = MyNewBertForTokenClassification.from_pretrained(best_name)
                model_to_save = (
                    model.module if hasattr(model, "module") else model
                ) 
                best_output_dir = os.path.join(output_dir,"best_model")
                if not os.path.exists(best_output_dir):
                     os.makedirs(best_output_dir)
                model_to_save.save_pretrained(best_output_dir)
                torch.save(saved_args, os.path.join(best_output_dir, "training_args.bin"))
            
        return results   
            
 


from senti_c import SentenceSentimentClassification

sentence_classifier = SentenceSentimentClassification()

test_data = ["我很喜歡這家店！超級無敵棒！","這個服務生很不親切..."]  
result = sentence_classifier.predict(test_data, run_split=True, aggregate_strategy=False)
print("\n *** \nBelow is the results:")
print(result)



# coding=utf-8
# Some of structures or codes in this script are referenced from HuggingFace Inc. team. 

# Copyright 2018 The Google AI Language Team Authors and The HuggingFace Inc. team.
# Copyright (c) 2018, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
'''屬性情感分類的資料讀取與處理'''

import os
import copy
import csv
import json
import logging
import numpy as np
import torch

logger = logging.getLogger(__name__)

ASPECT_LABEL_MAP = {0:"B", 1:"I", 2:"O"}
SENTIMENT_LABEL_MAP = {0:"NEG",1:"NEU",2:"POS",3:"O"}


class InputExample(object):
    """A single training/test example for token classification."""

    def __init__(self, guid, words, labels=None):
        """Constructs a InputExample.

        Args:
            guid: Unique id for the example.
            words: list. The words of the sequence.
            labels: (Optional) list. The labels for each word of the sequence. This should be
            specified for train and dev examples, but not for test examples.
        """
        self.guid = guid
        self.words = words
        self.labels = labels


class InputFeatures(object):
    """A single set of features of data."""

    def __init__(self, input_ids, attention_mask, token_type_ids, label=None,evaluate_label_ids=None):
        self.input_ids = input_ids
        self.attention_mask = attention_mask
        self.token_type_ids = token_type_ids
        self.label = label
        
        self.evaluate_label_ids = evaluate_label_ids 


class AspectProcessor:
    def _read_tsv(cls, input_file, quotechar=None):
        with open(input_file, "r", encoding="utf-8-sig") as f:
            return list(csv.reader(f, delimiter="\t", quotechar=quotechar))
        
    def get_train_dev_examples(self, data_dir, mode):
        file_path = os.path.join(data_dir, mode+".txt")
        return self._create_examples(file_path, mode)

    def get_test_examples_from_list(self, lines,finish_word_seg):
        '''data為list格式'''
        return self._create_examples_from_test(lines,finish_word_seg)
    
    def get_test_examples_from_tsv(self, data_dir):
        '''這裡的 data_dir 為使用者自己指定的檔名(路徑)'''
        
        ## 讀取tsv檔案
        data = self._read_tsv(data_dir) #得到一個大list，裡面每個list中包含tsv檔案中的一列資料
        
        ## 將資料整理成一個list,裡面每個元素都是一個String
        new_data = [i[0] for i in data] #因為只有文本所以取[0] 
        
        return new_data

    def get_labels(self):
        return ['B-NEG', 'B-NEU', 'B-POS', 'I-NEG', 'I-NEU', 'I-POS', 'O']   
    
    def _create_examples_from_test(self, lines, finish_word_seg):
        '''這是給沒有label的test資料所使用；lines型別為list'''
        
        examples = []
        all_data_words = [] #保存每句話對應的要預測的字
        for (i, line) in enumerate(lines):
            guid = "test-%s" % (i)
            
            if finish_word_seg: #使用者已用空白隔開各字，每個字就是要預測的內容
                text_a = line.split(" ")     
            else:
                text_a = list(line)
                
            text_a = [tmp for tmp in text_a if (len(tmp) > 0 and not tmp.isspace())] #過濾空白的字
            labels = [None] * len(text_a)
            
            all_data_words.append(text_a)
     
            examples.append(InputExample(guid=guid, words=text_a, labels=labels))
        return examples,all_data_words
    
    def _create_examples(self,file_path, mode):
        '''創建訓練和驗證時的資料examples'''
        
        guid_index = 0
        examples = []
        with open(file_path, encoding="utf-8") as f:
            words = []
            labels = []
            for line in f:
                if line.startswith("-DOCSTART-") or line == "" or line == "\n":
                    if words:
                        examples.append(InputExample(guid="{}-{}".format(mode, guid_index), words=words, labels=labels))
                        guid_index += 1
                        words = []
                        labels = []
                else:
                    splits = line.split(" ")
                    words.append(splits[0])
                    if len(splits) > 1:
                        labels.append(splits[-1].replace("\n", ""))

            if words:
                examples.append(InputExample(guid="{}-{}".format(mode, guid_index), words=words, labels=labels))
        return examples

    
    
def aspect_convert_examples_to_features(
    examples,
    tokenizer,
    label_list=None,
    max_length=128,
    pad_token=0,
    pad_token_segment_id=0,
    pad_token_label_id=-100,
    mask_padding_with_zero=True,
    logging_display=True,
):
    """
    Loads a data file into a list of ``InputFeatures``

    Returns:
        If the input is a list of ``InputExamples``, will return
        a list of task-specific ``InputFeatures`` which can be fed to the model.

    """
    
    label_map = {label: i for i, label in enumerate(label_list)}
# label map:  {'B-NEG': 0, 'B-NEU': 1, 'B-POS': 2, 'I-NEG': 3, 'I-NEU': 4, 'I-POS': 5, 'O': 6}

    features = []
    for (ex_index, example) in enumerate(examples):
        if ex_index % 10000 == 0 and logging_display:
            logger.info("Writing example %d of %d", ex_index, len(examples))

        tokens = []
        label_ids = []
        evaluate_label_ids = []  #該句子中實際需要去預測/計算的位置 (等於非特殊符號或subword的位置)
        valid_idx = 0 
        
        for word, label in zip(example.words, example.labels):
            word_tokens = tokenizer.tokenize(word)
            tokens.extend(word_tokens)
            if label is not None:
                label_ids.extend([label_map[label]] + [pad_token_label_id] * (len(word_tokens) - 1))
            else:
                label_ids.extend([None] + [pad_token_label_id] * (len(word_tokens) - 1))
                
            evaluate_label_ids.append(valid_idx)
            valid_idx += len(word_tokens)    
        
        assert valid_idx == len(tokens)
        evaluate_label_ids = np.array(evaluate_label_ids, dtype=np.int32)
        
        special_tokens_count = 2
        if len(tokens) > max_length - special_tokens_count:
            tokens = tokens[: (max_length - special_tokens_count)]
            label_ids = label_ids[: (max_length - special_tokens_count)]

            
        tokens += ["[SEP]"]
        label_ids += [pad_token_label_id]
        segment_ids = [0] * len(tokens)

        tokens = ["[CLS]"] + tokens
        label_ids = [pad_token_label_id] + label_ids
        segment_ids = [0] + segment_ids
        evaluate_label_ids += 1  #因為最前面加了"CLS"，所以索引都加1

        input_ids = tokenizer.convert_tokens_to_ids(tokens)
        
        input_mask = [1 if mask_padding_with_zero else 0] * len(input_ids)

        padding_length = max_length - len(input_ids)
        
        input_ids += [pad_token] * padding_length
        input_mask += [0 if mask_padding_with_zero else 1] * padding_length
        segment_ids += [pad_token_segment_id] * padding_length
        label_ids += [pad_token_label_id] * padding_length
        

        assert len(input_ids) == max_length
        assert len(input_mask) == max_length
        assert len(segment_ids) == max_length
        assert len(label_ids) == max_length

        if ex_index < 3 and logging_display:
            logger.info("*** Example ***")
            logger.info("guid: %s", example.guid)
            logger.info("tokens: %s", " ".join([str(x) for x in tokens]))
            logger.info("input_ids: %s", " ".join([str(x) for x in input_ids]))
            logger.info("input_mask: %s", " ".join([str(x) for x in input_mask]))
            logger.info("segment_ids: %s", " ".join([str(x) for x in segment_ids]))
            if label_ids[1] is not None:
                logger.info("label_ids: %s", " ".join([str(x) for x in label_ids]))

        features.append(
            InputFeatures(input_ids=input_ids, attention_mask=input_mask, token_type_ids=segment_ids, label=label_ids,evaluate_label_ids=evaluate_label_ids)
        )
    

    return features


def match_ot(gold_ote_sequence, pred_ote_sequence):
    
    n_hit = 0
    for t in pred_ote_sequence:
        if t in gold_ote_sequence:
            n_hit += 1
    return n_hit


def match_ts(gold_ts_sequence, pred_ts_sequence):

    tag2tagid = {'NEG': 0, 'NEU': 1, 'POS': 2}
    hit_count, gold_count, pred_count = np.zeros(3), np.zeros(3), np.zeros(3)
    
    for t in gold_ts_sequence:
        ts_tag = t[2]  #拿到該pair的情感
        tid = tag2tagid[ts_tag]
        gold_count[tid] += 1
    for t in pred_ts_sequence:
        ts_tag = t[2]
        tid = tag2tagid[ts_tag]
        if t in gold_ts_sequence:
            hit_count[tid] += 1
        pred_count[tid] += 1
        
    return hit_count, gold_count, pred_count


def tag2ot(ote_tag_sequence):
    '''從輸入的屬性 BIO tag 中取出每個屬性對應的開始與結束位置索引'''
    
    n_tags = len(ote_tag_sequence)
    ot_sequence = []
    beg, end = -1, -1
    checkafters = []  #每個位置會放該位置後面有幾個
    for i in range(n_tags):
        tag = ote_tag_sequence[i]
        
        if tag == 'B':  #只抓取從'B'開始的才納為屬性
            beg = i
            end = i  #預設為i，避免這個位置是整句話中最後一個位置
            
            for j in range(i+1,n_tags):   #從該indexs後面開始，如果i+1 = n_tags , 則不會進入此區間 (表示i為這句話中的最後一個位置)
                if ote_tag_sequence[j] == 'O' or ote_tag_sequence[j] == 'B':  #表示後面接O或B
                    end = j-1
                    break
                elif ote_tag_sequence[j] == 'I':
                    end = j  #先紀錄，避免這個位置(j)為整句話中最後一個字
            
            if end >= beg > -1:
                ot_sequence.append((beg, end))
                beg, end = -1, -1
            
    return ot_sequence


def tag2ts(ts_tag_sequence):
    '''從輸入的屬性-情感 tag 中取出每個屬性對應的開始與結束位置索引與情感，注意只有同一屬性中每個字情感皆一致才提取'''
    
    n_tags = len(ts_tag_sequence)
    ts_sequence, sentiments = [], []
    beg, end = -1, -1
    
    for i in range(n_tags):
        ts_tag = ts_tag_sequence[i]
        
        # current position and sentiment
        eles = ts_tag.split('-')

        pos, sentiment = eles
        
        if pos == 'B':  #抓取從'B'開始的
            beg = i
            end = i  #預設為i，避免這個位置是整句話中最後一個位置
            sentiments.append(sentiment)  #將其對應情感加入
            
            for j in range(i+1,n_tags):   #從該indexs後面開始，如果i+1 = n_tags , 則不會進入此區間 (表示i為這句話中的最後一個位置)
                tmp = ts_tag_sequence[j].split('-')
                
                if tmp[0] == 'B' or tmp[0] == 'O':  #表示後面接O或B
                    end = j-1
                    break
                elif tmp[0] == 'I': 
                    sentiments.append(tmp[1])
                    end = j  #先紀錄，避免這個位置(j)為整句話中最後一個字
    
            # 該屬性中每個字的情感都一致時才會納入
            # 並避免遇到屬性有預測(有找到(start,end))、但情感部分預測為O (有此狀況則不納入)
            if end >= beg > -1 and len(set(sentiments)) == 1 and sentiments[0] != 'O':  
                ts_sequence.append((beg, end, sentiment))
                sentiments = []
                beg, end = -1, -1
    
    return ts_sequence

def evaluate_ote(gold_ot, pred_ot):
    """
    evaluate the model performance for the ote task
    :param gold_ot: gold standard ote tags 
    :param pred_ot: predicted ote tags
    """
    
    assert len(gold_ot) == len(pred_ot)
    
    n_samples = len(gold_ot)
    
    n_tp_ot, n_gold_ot, n_pred_ot = 0, 0, 0
    
    for i in range(n_samples):  
        g_ot = gold_ot[i]  #對每筆資料來做
        p_ot = pred_ot[i]
      
        g_ot_sequence, p_ot_sequence = tag2ot(ote_tag_sequence=g_ot), tag2ot(ote_tag_sequence=p_ot)

        
        # 計算 hit number
        n_hit_ot = match_ot(gold_ote_sequence=g_ot_sequence, pred_ote_sequence=p_ot_sequence)
        n_tp_ot += n_hit_ot
        n_gold_ot += len(g_ot_sequence)
        n_pred_ot += len(p_ot_sequence)
    
    ot_precision = float(n_tp_ot) / float(n_pred_ot + 0.001)
    ot_recall = float(n_tp_ot) / float(n_gold_ot + 0.001)
    ot_f1 = 2 * ot_precision * ot_recall / (ot_precision + ot_recall + 0.001)
    
    ote_scores = {"aspect_precision": ot_precision, 
                  "aspect_recall": ot_recall, 
                  "aspect_f1": ot_f1}
    
    return ote_scores


def evaluate_ts(gold_ts, pred_ts):
    """
    evaluate the model performance for the ts task
    :param gold_ts: gold standard ts tags
    :param pred_ts: predicted ts tags
    """
    
    assert len(gold_ts) == len(pred_ts)
    n_samples = len(gold_ts)

    n_tp_ts, n_gold_ts, n_pred_ts = np.zeros(3), np.zeros(3), np.zeros(3)
    ts_precision, ts_recall, ts_f1 = np.zeros(3), np.zeros(3), np.zeros(3)

    for i in range(n_samples):
        g_ts = gold_ts[i]
        p_ts = pred_ts[i]
     
        g_ts_sequence, p_ts_sequence = tag2ts(ts_tag_sequence=g_ts), tag2ts(ts_tag_sequence=p_ts)
        
        hit_ts_count, gold_ts_count, pred_ts_count = match_ts(gold_ts_sequence=g_ts_sequence,
                                                              pred_ts_sequence=p_ts_sequence)
        
        n_tp_ts += hit_ts_count
        n_gold_ts += gold_ts_count
        n_pred_ts += pred_ts_count
        
    for i in range(3):  #有三類別
        n_ts = n_tp_ts[i]
        n_g_ts = n_gold_ts[i]
        n_p_ts = n_pred_ts[i]
        
        ts_precision[i] = float(n_ts) / float(n_p_ts + 0.001)
        ts_recall[i] = float(n_ts) / float(n_g_ts + 0.001)
        ts_f1[i] = 2 * ts_precision[i] * ts_recall[i] / (ts_precision[i] + ts_recall[i] + 0.001)

    ts_macro_f1 = ts_f1.mean()  #算macro f1

    n_tp_total = sum(n_tp_ts)
    n_g_total = sum(n_gold_ts)
    n_p_total = sum(n_pred_ts)

    ts_micro_p = float(n_tp_total) / (n_p_total + 0.001)
    ts_micro_r = float(n_tp_total) / (n_g_total + 0.001)
    ts_micro_f1 = 2 * ts_micro_p * ts_micro_r / (ts_micro_p + ts_micro_r + 0.001)
    
    ts_scores = {"absa_macro_neg": ts_f1[0],
                 "absa_macro_neu": ts_f1[1],
                 "absa_macro_pos": ts_f1[2], 
                 "absa_macro_f1": ts_macro_f1,
                 "absa_micro_precision": ts_micro_p, 
                 "absa_micro_recall": ts_micro_r, 
                 "absa_micro_f1": ts_micro_f1}
    
    return ts_scores
                                    


def chg_labels_to_aspect_and_sentiment(input_labels):
    '''將原本合併的屬性-情感標記轉成分開的形式'''
    
    aspect_labels =  input_labels.clone().detach()  
    sentiment_labels =  input_labels.clone().detach()
    
    for i,j in enumerate(input_labels):  #每個j都是該batch中的一筆資料
        for m,n in enumerate(j):  #每個n都是一筆資料中的一個token
            if n == 0:  # 'B-NEG'
                aspect_labels[i][m] = 0
                sentiment_labels[i][m] = 0
            if n == 1:  # 'B-NEU'
                aspect_labels[i][m] = 0
                sentiment_labels[i][m] = 1    
            if n == 2:  # 'B-POS'
                aspect_labels[i][m] = 0
                sentiment_labels[i][m] = 2
            if n == 3:  # 'I-NEG'
                aspect_labels[i][m] = 1
                sentiment_labels[i][m] = 0   
            if n == 4:  # 'I-NEU'
                aspect_labels[i][m] = 1
                sentiment_labels[i][m] = 1
            if n == 5:  # 'I-POS'
                aspect_labels[i][m] = 1
                sentiment_labels[i][m] = 2
            if n == 6:  # 'O'
                aspect_labels[i][m] = 2
                sentiment_labels[i][m] = 3
            if n == -100 :  
                aspect_labels[i][m] = -100
                sentiment_labels[i][m] = -100        
    
    return aspect_labels,sentiment_labels


# coding=utf-8
# Some of structures or codes in this script are referenced from HuggingFace Inc. team. 

# Copyright 2018 The Google AI Language Team Authors and The HuggingFace Inc. team.
# Copyright (c) 2018, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
'''屬性情感分類的預測相關程式'''

import logging
import os,sys
import numpy as np
import torch
from torch.utils.data import DataLoader, SequentialSampler, TensorDataset , DistributedSampler
import json
from tqdm import tqdm
from pathlib import Path

from transformers import (
    BertTokenizer,  
)

from .model_structure import MyNewBertForTokenClassification  
from .utils import AspectProcessor, aspect_convert_examples_to_features as convert_examples_to_features, tag2ot, tag2ts, evaluate_ote, evaluate_ts, ASPECT_LABEL_MAP, SENTIMENT_LABEL_MAP,split_text_from_input_data,get_toolkit_models


logger = logging.getLogger(__name__) 

ASPECT_CLASSIFICATION_MODEL = {
    'default' : '/pretrained_model/chinese_aspect_model',  #以本研究繼續預訓練模型進行微調後的最佳模型
    'open' : '/pretrained_model/chinese_aspect_model'      #以開源預訓練模型進行微調後的最佳模型  
}  

DEFAULT_ASPECT_CLASSIFICATION_TOKENZIER = 'bert-base-chinese' 


class AspectSentimentAnalysis:
    '''屬性情感分析的主要類別：使用提供的微調模型來預測給定句子的情感'''  
    
    def __init__(self,
                 model_name_or_path = 'default',
                 tokenizer = DEFAULT_ASPECT_CLASSIFICATION_TOKENZIER,
                 model_cache_folder_name = 'original',
                 no_cuda = False,
                 local_rank = -1,
                 logging_level = "info"
                ):
        '''
        設定屬性情感分析相關參數
        model_name_or_path: 用來預測屬性情感的模型，可以為放置模型的路徑、或是此工具所提供的模型名稱；默認為本工具提供的由繼續預訓練模型微調完畢的模型(參數值為"default")，如果想使用由開源預訓練模型所微調出的模型，參數值設為"open"
        tokenizer: 模型使用的tokenizer名稱，默認為"bert-base-chinese" 
        model_cache_folder_name: 存放本研究所開源模型的資料夾名稱，默認名稱為"original"，表示模型下載後存放於/pretrained_model/chinese_aspect_model/original資料夾中
        no_cuda: 是否避免使用gpu，預設"False"
        local_rank: 是否使用平行化運算，預設"-1"
        logging_level: 設置logging level，預設為"info"
        '''
        
        self.local_rank = local_rank
        
        llevel = logging.WARN
        self.logging_display = False
        if logging_level == "info":
            llevel = logging.INFO
            self.logging_display = True
        
        ### 設定 logging ###
        
        logging.basicConfig(
            format="%(asctime)s - %(levelname)s - %(name)s -   %(message)s",
            datefmt="%m/%d/%Y %H:%M:%S",
            # slevel=logging.INFO if self.local_rank in [-1, 0] else logging.WARN,
            level = llevel
        )   
        
        ### 匯入 tokenizer,model ###
        hf_cache_dir = str(Path.home()) + "/senti_c/huggingface/"
        if not os.path.exists(hf_cache_dir):
            os.makedirs(hf_cache_dir)
        
        
        self.tokenizer = BertTokenizer.from_pretrained(tokenizer, cache_dir = hf_cache_dir)
        
        if model_name_or_path in ASPECT_CLASSIFICATION_MODEL:
           ## 下載與解壓縮預設模型並存放於與預設模型同一資料夾 ##
            model_paths = os.path.split(os.path.realpath(__file__))[0] + ASPECT_CLASSIFICATION_MODEL[model_name_or_path] + "/" + model_cache_folder_name #取得模型資料夾位置絕對路徑
            
            if not os.path.exists(model_paths):
                createok = False
                try:
                    os.makedirs(model_paths)
                    createok = True
                except Exception as e:
                    logger.warning(f"Encountered error when creating folder {model_paths}: {e}")
                
                if createok == False:                    
                    # tmpdir = tempfile.TemporaryDirectory()
                    # print(f"tmpdir = {tmpdir.name}")                    
                    homedir = str(Path.home())
                    # print(f"homedir = {homedir}")  
                    model_paths = homedir + "/senti_c/" + ASPECT_CLASSIFICATION_MODEL[model_name_or_path] + "/" + model_cache_folder_name
                    logger.info(f"Try to create model folder at temp dir: {model_paths}")
                    os.makedirs(model_paths)
                
            get_toolkit_models(model_paths, "aspect",model_name_or_path)
            self.model = MyNewBertForTokenClassification.from_pretrained(model_paths)
        else:
            self.model = MyNewBertForTokenClassification.from_pretrained(model_name_or_path)

        
        ### 設定 cpu,gpu ###
        
        if self.local_rank == -1 or no_cuda:
            self.device = torch.device("cuda" if torch.cuda.is_available() and not no_cuda else "cpu")
            self.n_gpu = 0 if no_cuda else torch.cuda.device_count()
        else:
            torch.cuda.set_device(self.local_rank)
            self.device = torch.device("cuda", self.local_rank)
            torch.distributed.init_process_group(backend="nccl")
            self.n_gpu = 1
            
        self.model.to(self.device)

        


    def _check_input_list(self,input_data_or_path):                                    
        '''檢查輸入資料(list)的格式'''
        
        data = []
        if type(input_data_or_path) != list:
            raise Exception("輸入資料型別錯誤！請重新確認！")
        elif len(input_data_or_path) == 0 :
            raise Exception("輸入資料大小為0！請重新確認！")
        else:
            data = input_data_or_path 
        return data

    
    def _check_input_tsv(self,input_data_or_path):                                    
        '''檢查輸入資料(tsv)的格式；並將tsv資料轉成list'''
        
        data = []
        
        try:
            if not os.path.isfile(input_data_or_path):
                raise Exception        
        except:
            raise Exception("輸入資料不存在或並非檔案格式！請重新確認！")
            
        processor = AspectProcessor()
        data = processor.get_test_examples_from_tsv(input_data_or_path)
        return data 
    

    def _read_input_list(self,input_data,output_dir=None,finish_word_seg=False,overwrite_cache=True):                             
        '''產生符合模型需求的資料格式'''
        
        ## 獲取examples和每句話要預測的字 ##
        processor = AspectProcessor()
        examples,all_data_words = processor.get_test_examples_from_list(input_data,finish_word_seg)
        
        if output_dir is not None: #表示有指定輸出路徑
            cached_features_file = os.path.join(
                output_dir,
                "cached_test",
            )
        else: #將路徑改成當前目錄
            cached_features_file = os.path.join(
                ".",
                "cached_test",
            )
        
        if os.path.exists(cached_features_file) and not overwrite_cache:
            logger.info("載入先前的 cache 檔案 %s", cached_features_file)
            features = torch.load(cached_features_file)
        else:    
            logger.info("創建 cache 檔案 %s", cached_features_file)
                
            ## 將資料轉成特徵 ##
            labels=processor.get_labels()
            features = convert_examples_to_features(
                examples,
                self.tokenizer,
                label_list=labels,
                max_length=128,
                pad_token=self.tokenizer.convert_tokens_to_ids([self.tokenizer.pad_token])[0],
                logging_display=self.logging_display,
            )  
            
            
            if self.local_rank in [-1, 0]:
                logger.info("儲存 cache 檔案 %s", cached_features_file)
                torch.save(features, cached_features_file)
        
        ## 將特徵轉成tensors並建成Dataset ##
        all_input_ids = torch.tensor([f.input_ids for f in features], dtype=torch.long)
        all_attention_mask = torch.tensor([f.attention_mask for f in features], dtype=torch.long)
        all_token_type_ids = torch.tensor([f.token_type_ids for f in features], dtype=torch.long)
        
        all_evaluate_label_ids = [f.evaluate_label_ids for f in features]  #驗證或預測時會用到
        
        dataset = TensorDataset(all_input_ids, all_attention_mask, all_token_type_ids)
        
        return dataset , all_evaluate_label_ids , all_data_words
            
    
    def predict(self,
                input_data_or_path,
                input_mode="list",
                finish_word_seg=False,
                run_split=False,
                batch_size=8,
                output_mode="return_result",
                output_dir=None,          
                output_result="only_extraction",
                overwrite_cache=True,
               ):
        '''
        預測屬性與屬性情感的專用函數。(注意初始化class後，可傳入不同參數來重複呼叫此函數進行預測)
        參數包含：
        input_data_or_path : 輸入數據，和input_mode相關聯與對應，默認為輸入list型別  
        input_mode : 輸入模式選擇，分為"list"或"tsv"，默認"list"
        finish_word_seg : 每筆輸入數據中的每個字是否已經用空白隔開，默認"False"。如果為"True"，則程式處理輸入數據時會直接以split將每個字切開、並對切開後各字做屬性與情感預測；如果為"False"，則程式處理輸入數據時會直接以list方式來將句子內容切成字、並對這種方式所切出的各字做預測；通常如果輸入數據夾雜中英文，則建議輸入數據前先以空白隔開每個要標記的字、避免程式後續處理時標記範圍錯誤，至於如果輸入內容為全中文則直接以原始文本輸入即可
        run_split : 是否對每筆輸入資料執行斷句，默認"False"
        batch_size : 一次丟入多筆待預測資料到模型，默認"8"
        output_mode : 輸出模式選擇，分為"寫入檔案(json檔)+回傳預測變數(write_file)"或"單純回傳預測變數(return_result)"，默認"單純回傳預測變數(return_result)"
        output_dir : 輸出檔案路徑，可選參數
        output_result : 輸出內容選擇，分為"從標記結果提取屬性與情感(only_extraction)"或"原始序列標記結果(only_tags)"或"原始序列標記與提取後的結果皆回傳(all)"，默認為"從標記結果提取出屬性與情感(only_extraction)"
        overwrite_cache : 是否覆寫預測資料的cache檔案，默認為True，若要重複預測同樣的內容，可將此參數改為False，程式會讀取先前儲存的cache檔案
        '''

        
        all_vals = locals()
        del all_vals["self"]  #不須顯示此項
        logger.info("predict函數的所有參數： %s",all_vals)
        logger.info("開始讀取與檢查輸入數據！")
        
        if not input_data_or_path: #如果沒有任何值
            raise Exception("缺少輸入資料！請重新確認！")
        
        if input_mode == 'list':
            ### 檢查輸入資料格式 ###
            input_data = self._check_input_list(input_data_or_path)
        else:
            ### 檢查輸入資料格式，並獲得list格式的資料 ###
            input_data = self._check_input_tsv(input_data_or_path)    

            
        ## 依序處理每筆資料 ##
        all_data = [] #儲存所有要丟入模型的資料
        input_data_chg = []  #儲存每個句子對應的原始文本內容，這是給後面輸出時所用
        split_index = [] #儲存原本輸入文本所對應的累積句子數 (因為當輸入文本經過斷句後，可能一個輸入文本包含多個句子，為了方便後續作匯總情感，所以儲存這項) => e.g. 輸入文本有三個，斷句後分別有1,3,2個句子，則對應的 split_index 為[1,4,6]，到時要辨別的話，便是用0:1,1:4,4:6來取得
        current_counts = 0 #目前累積句子數
        
        for data in input_data: 
            ## 執行斷句 ##
            if run_split:
                sentences = split_text_from_input_data(data)  
            else:
                sentences = [data]
            
            sentences = [line for line in sentences if (len(line) > 0 and not line.isspace())] #去除整個句子中為空白的句子 
              
            all_data.extend(sentences)
            current_counts += len(sentences)
            split_index.append(current_counts)

            input_data_chg.extend([data] * len(sentences))
        
        try:
            assert len(all_data) == current_counts == len(input_data_chg)
        except:
            raise Exception("處理過程中發生錯誤！請聯繫開發者！")
        
        
        logger.info("原本的總輸入文本數目：%s",len(input_data))
        logger.info("處理後的總文本/句子數目：%s",current_counts)
        
        
        ## 將資料轉成模型需要的dataset & 獲取相關必要變數 ##  
        logger.info("開始處理輸入資料為模型需求格式！")
        test_dataset, all_evaluate_label_ids, all_data_words = self._read_input_list(all_data,output_dir,finish_word_seg, overwrite_cache)
        
        try:
            assert len(all_data) == len(all_evaluate_label_ids) == len(all_data_words)
        except:
            raise Exception("處理過程中發生錯誤！請聯繫開發者！")
        
            
        ## 實際跑模型獲得預測結果 ##
        test_batch_size = batch_size * max(1, self.n_gpu)
        test_sampler = SequentialSampler(test_dataset) if self.local_rank == -1 else DistributedSampler(test_dataset)
        test_dataloader = DataLoader(test_dataset, sampler=test_sampler, batch_size=test_batch_size)
        
        if self.n_gpu > 1:
            self.model = torch.nn.DataParallel(self.model)
        
        
        logger.info("***** 開始進行預測  *****")
        logger.info("  資料數量 = %d", len(test_dataset))
        logger.info("  Batch大小 = %d", test_batch_size)
        
        
        aspect_preds = None
        sentiment_preds = None
        
        for batch in tqdm(test_dataloader, desc="Predicting"):
            self.model.eval()
            batch = tuple(t.to(self.device) for t in batch)

            with torch.no_grad():
                inputs = {"input_ids": batch[0], "attention_mask": batch[1],"token_type_ids": batch[2]} #因為要預測，所以沒有labels
 
                outputs = self.model(**inputs)
                aspect_logits, sentiment_logits = outputs[:2]  #取得預測屬性和預測情感標記 (注意沒有labels所以不會回傳loss)   
            
            if aspect_preds is None: 
                aspect_preds = aspect_logits.detach().cpu().numpy()
                sentiment_preds = sentiment_logits.detach().cpu().numpy()
            else:
                aspect_preds = np.append(aspect_preds, aspect_logits.detach().cpu().numpy(), axis=0)
                sentiment_preds = np.append(sentiment_preds, sentiment_logits.detach().cpu().numpy(), axis=0)
        
        
        logger.info("已完成預測！開始整理預測結果！")
        
        ## 整理預測值(標記) ##
        aspect_preds = np.argmax(aspect_preds, axis=2)
        sentiment_preds = np.argmax(sentiment_preds, axis=2)
        
        aspect_preds_list = [[] for _ in range(aspect_preds.shape[0])]
        sentiment_preds_list = [[] for _ in range(sentiment_preds.shape[0])]
        absa_preds_list = [[] for _ in range(sentiment_preds.shape[0])]
        

        for i in range(aspect_preds.shape[0]): #句子總數
            evaluate_label_ids = all_evaluate_label_ids[i] #取得該句子中有效的位置索引
            for j in range(aspect_preds.shape[1]): #句子長度
                if j in evaluate_label_ids: 
                    asp_pred = ASPECT_LABEL_MAP[aspect_preds[i][j]]
                    sent_pred = SENTIMENT_LABEL_MAP[sentiment_preds[i][j]]

                    aspect_preds_list[i].append(asp_pred)
                    sentiment_preds_list[i].append(sent_pred)

                    # 將屬性標記和情感標記標籤結合
                    absa_preds_list[i].append(asp_pred+"-"+sent_pred)
        
    
        ## 從標記結果提取屬性/情感 (不會單獨取出情感) ##
        pred_aspect_extraction = [] #每個元素為一個list，會放置該句子提取出的屬性術語，如果該句子沒有預測任何屬性，則為空的list
        pred_absa_extraction = [] #每個元素為一個list，會放置該句子提取出的屬性術語+對應情感(tuple形式，如：("服務","POS"))
        for aspect_list,absa_list,text_list in zip(aspect_preds_list,absa_preds_list,all_data_words):  #對每個句子
            
            ot_sequence = tag2ot(aspect_list)
            ts_sequence = tag2ts(absa_list)
            
            cur_aspect = []
            for beg,end in ot_sequence:
                cur_aspect.append("".join(text_list[beg:end+1]))
            pred_aspect_extraction.append(cur_aspect)    
            
            cur_absa = []
            for beg,end,sentiment in ts_sequence:
                cur_absa.append(("".join(text_list[beg:end+1]),sentiment))
                
            pred_absa_extraction.append(cur_absa)    
            
        
        ## 整理要輸出或返回的預測結果 ##
        if output_result == 'only_extraction':
            # 每個元素都是一個list中包含多個list，每個子層list對應一個句子的字/提取出的結果
            pred_result = {"InputWords":all_data_words,"AspectTermExtraction":pred_aspect_extraction, "AspectTermAndSentimentExtraction":pred_absa_extraction}
            
        elif output_result == 'only_tags':
            # 每個元素都是一個list中包含多個list，每個子層list對應一個句子的字/預測的標記
            pred_result = {"InputWords":all_data_words ,"AspectTermTags":aspect_preds_list, "SentimentTags":sentiment_preds_list, "AspectTermAndSentimentTags":absa_preds_list}
            
        else:
            # 每個元素都是一個list中包含多個list，每個子層list對應一個句子的字/提取出的結果/預測標記
            pred_result = {"InputWords":all_data_words,"AspectTermExtraction":pred_aspect_extraction, "AspectTermAndSentimentExtraction":pred_absa_extraction, "AspectTermTags":aspect_preds_list, "SentimentTags":sentiment_preds_list, "AspectTermAndSentimentTags":absa_preds_list}

        # 額外新增各句子對應的原始文本，方便斷句版本的做比較與對應，其為一個list，list中包含每個輸入文本(string) 
        pred_result["OriInputs"] = input_data_chg
        
        
        ## 返回/寫入預測結果 ##    
        if output_mode == 'write_file':
            if output_dir is not None:
                output_test_predictions_file = os.path.join(output_dir,"predictions.json")
            else:
                output_test_predictions_file = "predictions.json"  #表示會放在與目前位置同一路徑
                
            logger.info("將預測結果寫入檔案 %s",output_test_predictions_file)
            
            # 以json方式寫入檔案 #
            with open(output_test_predictions_file, 'w') as f:
                json.dump(pred_result, f , ensure_ascii=False) #加入 ensure_ascii=False 避免亂碼 

        return pred_result  #不管哪種模式都會回傳預測結果的變數，使用者可以依據需求選擇是否使用 
        
        
        
            
            
        
          


# coding=utf-8
# Some of structures or codes in this script are referenced from HuggingFace Inc. team. 

# Copyright 2018 The Google AI Language Team Authors and The HuggingFace Inc. team.
# Copyright (c) 2018, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
'''句子情感分類的資料讀取與處理'''

import os
import copy
import csv
import json
import logging
import numpy as np

logger = logging.getLogger(__name__)

class InputExample(object):
    """
    A single training/test example for simple sequence classification.

    Args:
        guid: Unique id for the example.
        text_a: string. The untokenized text of the first sequence. For single
            sequence tasks, only this sequence must be specified.
        text_b: (Optional) string. The untokenized text of the second sequence.
            Only must be specified for sequence pair tasks.
        label: (Optional) string. The label of the example. This should be
            specified for train and dev examples, but not for test examples.
    """

    def __init__(self, guid, text_a, text_b=None, label=None):
        self.guid = guid
        self.text_a = text_a
        self.text_b = text_b
        self.label = label

    def __repr__(self):
        return str(self.to_json_string())

    def to_dict(self):
        """Serializes this instance to a Python dictionary."""
        output = copy.deepcopy(self.__dict__)
        return output

    def to_json_string(self):
        """Serializes this instance to a JSON string."""
        return json.dumps(self.to_dict(), indent=2, sort_keys=True) + "\n"


class InputFeatures(object):
    """
    A single set of features of data.

    Args:
        input_ids: Indices of input sequence tokens in the vocabulary.
        attention_mask: Mask to avoid performing attention on padding token indices.
            Mask values selected in ``[0, 1]``:
            Usually  ``1`` for tokens that are NOT MASKED, ``0`` for MASKED (padded) tokens.
        token_type_ids: Segment token indices to indicate first and second portions of the inputs.
        label: Label corresponding to the input
    """

    def __init__(self, input_ids, attention_mask=None, token_type_ids=None, label=None):
        self.input_ids = input_ids
        self.attention_mask = attention_mask
        self.token_type_ids = token_type_ids
        self.label = label

    def __repr__(self):
        return str(self.to_json_string())

    def to_dict(self):
        """Serializes this instance to a Python dictionary."""
        output = copy.deepcopy(self.__dict__)
        return output

    def to_json_string(self):
        """Serializes this instance to a JSON string."""
        return json.dumps(self.to_dict(), indent=2, sort_keys=True) + "\n"


class SentenceProcessor:
    def _read_tsv(cls, input_file, quotechar=None):
        with open(input_file, "r", encoding="utf-8-sig") as f:
            return list(csv.reader(f, delimiter="\t", quotechar=quotechar)) 
        
    def get_train_dev_examples(self, data_dir, mode):
        return self._create_examples(self._read_tsv(os.path.join(data_dir, mode+".tsv")), mode)
    
    def get_test_examples_from_list(self, lines):
        '''data為list格式'''
        return self._create_examples_from_test(lines)
    
    def get_test_examples_from_tsv(self, data_dir):
        '''這裡的 data_dir 為使用者自己指定的檔名(路徑)'''
        
        ## 讀取tsv檔案
        data = self._read_tsv(data_dir) #得到一個大list，裡面每個list中包含tsv檔案中的一列資料
        
        ## 將資料整理成一個list,裡面每個元素都是一個String
        new_data = [i[0] for i in data] #因為只有文本所以取[0] 
        
        return new_data

    def get_labels(self):
        return ["0", "1"]   

    def _create_examples(self, lines, set_type):
        '''創建訓練和驗證時的資料examples'''
        
        examples = []
        for (i, line) in enumerate(lines):
            guid = "%s-%s" % (set_type, i)
            text_a = line[0]
            label = line[1:]   
            examples.append(InputExample(guid=guid, text_a=text_a, text_b=None, label=label))
        return examples
    
    def _create_examples_from_test(self, lines):
        '''這是給沒有label的test資料所使用；lines型別為list'''
        
        examples = []
        for (i, line) in enumerate(lines):
            guid = "test-%s" % (i)
            text_a = line
     
            examples.append(InputExample(guid=guid, text_a=text_a, text_b=None, label=None))
        return examples
    
    
def sentence_convert_examples_to_features(
    examples,
    tokenizer,
    label_list=None,
    max_length=256,
    pad_token=0,
    pad_token_segment_id=0,
    mask_padding_with_zero=True,
    logging_display=True,
):
    """
    Loads a data file into a list of ``InputFeatures``

    Returns:
        If the input is a list of ``InputExamples``, will return
        a list of task-specific ``InputFeatures`` which can be fed to the model.

    """
    label_map = {label: i for i, label in enumerate(label_list)}
    
    features = []
    for (ex_index, example) in enumerate(examples):
        len_examples = 0
        len_examples = len(examples)
        
        if ex_index % 10000 == 0 and logging_display:
            logger.info("Writing example %d/%d" % (ex_index, len_examples))

        inputs = tokenizer.encode_plus(
            example.text_a, example.text_b, add_special_tokens=True, max_length=max_length, return_token_type_ids=True,
        )
        input_ids, token_type_ids = inputs["input_ids"], inputs["token_type_ids"]

        # The mask has 1 for real tokens and 0 for padding tokens. Only real
        # tokens are attended to.
        attention_mask = [1 if mask_padding_with_zero else 0] * len(input_ids)

        # Zero-pad up to the sequence length.
        padding_length = max_length - len(input_ids)

        input_ids = input_ids + ([pad_token] * padding_length)
        attention_mask = attention_mask + ([0 if mask_padding_with_zero else 1] * padding_length)
        token_type_ids = token_type_ids + ([pad_token_segment_id] * padding_length)

        
        assert len(input_ids) == max_length, "Error with input length {} vs {}".format(len(input_ids), max_length)
        assert len(attention_mask) == max_length, "Error with input length {} vs {}".format(
            len(attention_mask), max_length
        )
        assert len(token_type_ids) == max_length, "Error with input length {} vs {}".format(
            len(token_type_ids), max_length
        )


        labels_ids = []
        if example.label is not None: #表示並非預測時，因為預測時這個會是none
            for label in example.label: 
                labels_ids.append(float(label))
            
            
        if ex_index < 3 and logging_display:
            logger.info("*** Example ***")
            logger.info("guid: %s" % (example.guid))
            logger.info("input_ids: %s" % " ".join([str(x) for x in input_ids]))
            logger.info("attention_mask: %s" % " ".join([str(x) for x in attention_mask]))
            logger.info("token_type_ids: %s" % " ".join([str(x) for x in token_type_ids]))
            if example.label is not None:
                logger.info("label: %s (id = %s)" % (example.label, (labels_ids,)))   

        features.append(
            InputFeatures(
                input_ids=input_ids, attention_mask=attention_mask, token_type_ids=token_type_ids, label=labels_ids
            )   
        )

    return features


def sigmoid_array(x):                                       
    return 1 / (1 + np.exp(-x))


import setuptools

with open("README.md", "r",encoding="utf-8") as f:
    long_description = f.read()

setuptools.setup(
    name="senti_c", 
    version="0.1.5",
    keywords='traditional chinese sentiment analysis tool',
    author="Julie Tu and Hsin-min Lu",
    author_email="luim@ntu.edu.tw",
    description="Traditional Chinese sentiment analysis tool based on BERT.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/hsinmin/senti_c",
    license='MIT',
    classifiers=[
        "Programming Language :: Python :: 3.8",
        "License :: OSI Approved :: MIT License",
        'Topic :: Software Development :: Build Tools',
    ],
    package_dir={'': 'src'},
    packages=setuptools.find_packages(where='src'),
    python_requires='>=3.7',
    install_requires=[
        'transformers==2.11.0',
        'scipy==1.4.1',
        'pandas',
        'scikit-learn',
        'torch>1.0,<1.9.9',
        'tensorflow==2.2.0',
    ],
)



# coding=utf-8
# Some of structures or codes in this script are referenced from HuggingFace Inc. team. 

# Copyright 2018 The Google AI Language Team Authors and The HuggingFace Inc. team.
# Copyright (c) 2018, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
'''放置句子情感分類、屬性情感分析模型架構'''   

import torch
import torch.nn as nn
from torch.nn import BCEWithLogitsLoss
from torch.nn import CrossEntropyLoss
from torch.nn import NLLLoss
from transformers import (
    BertModel, 
    BertPreTrainedModel,
)    
import numpy as np
import os
import torch.nn.functional as F 

from .utils import chg_labels_to_aspect_and_sentiment,get_domain_embedding

class MyNewBertForSequenceClassification(BertPreTrainedModel):
    '''BERT輸出接上一層線性層；此外額外加入domain embedding layer，等於最後BERT輸出和domain embedding連接、饋入分類層，是句子情感分類的模型'''
    
    def __init__(self, config):
        super().__init__(config)
        self.num_labels = config.num_labels

        self.bert = BertModel(config)
        self.dropout = nn.Dropout(config.hidden_dropout_prob)
        
        ### 加入 domain embedding layer
        file_path = os.path.split(os.path.realpath(__file__))[0] + "/utils/new_restaurant_emb.vec.npy"
        get_domain_embedding(file_path)
        domain_emb = np.load(file_path) 
        self.domain_embedding = nn.Embedding(domain_emb.shape[0], domain_emb.shape[1])
        self.domain_embedding.weight = nn.Parameter(torch.from_numpy(domain_emb), requires_grad=True)  
        
        self.first_classifier = nn.Linear(config.hidden_size+domain_emb.shape[1],config.hidden_size)
        self.classifier = nn.Linear(config.hidden_size, self.config.num_labels)  

        self.init_weights()

    def forward(
        self,
        input_ids=None,
        attention_mask=None,
        token_type_ids=None,
        position_ids=None,
        head_mask=None,
        inputs_embeds=None,
        labels=None,
    ):
        outputs = self.bert(
            input_ids,
            attention_mask=attention_mask,
            token_type_ids=token_type_ids,
            position_ids=position_ids,
            head_mask=head_mask,
            inputs_embeds=inputs_embeds,
        )

        pooled_output = outputs[1]

        pooled_output = self.dropout(pooled_output)
        
        ### 加入 domain embedding ### 
        input_embedding = self.domain_embedding(input_ids) 
    

        ### 將每筆資料中每個字的embedding拼起來平均 ###
        sent_embedding = torch.mean(input_embedding,dim=1)  #針對seq部分做平均
        
        ### 將平均後的embedding和bert的CLS位置輸出連接 ###
        combine_emb = torch.cat((sent_embedding,pooled_output),dim=1)

        ### 經過兩層線性層、中間添加tanh活化函數 ###
        firsts = self.first_classifier(combine_emb)
        firsts = torch.tanh(firsts)
        logits = self.classifier(firsts)

        outputs = (logits,) + outputs[2:]  

        if labels is not None:
            loss_fct = BCEWithLogitsLoss()  #改成多標籤分類(每個標籤做二元分類的感覺)
            labels = labels.float()  
            loss = loss_fct(logits.view(-1, self.num_labels), labels.view(-1, self.num_labels))

            outputs = (loss,) + outputs

        return outputs  


# class MyNewBertForSequenceClassification(BertPreTrainedModel):
#     '''BERT輸出接上一層線性層，是句子情感分類的模型'''
    
#     def __init__(self, config):
#         super().__init__(config)
#         self.num_labels = config.num_labels

#         self.bert = BertModel(config)
#         self.dropout = nn.Dropout(config.hidden_dropout_prob)
#         self.classifier = nn.Linear(config.hidden_size, self.config.num_labels)

#         self.init_weights()

#     def forward(
#         self,
#         input_ids=None,
#         attention_mask=None,
#         token_type_ids=None,
#         position_ids=None,
#         head_mask=None,
#         inputs_embeds=None,
#         labels=None,
#     ):
#         outputs = self.bert(
#             input_ids,
#             attention_mask=attention_mask,
#             token_type_ids=token_type_ids,
#             position_ids=position_ids,
#             head_mask=head_mask,
#             inputs_embeds=inputs_embeds,
#         )

#         pooled_output = outputs[1]

#         pooled_output = self.dropout(pooled_output)
#         logits = self.classifier(pooled_output)

#         outputs = (logits,) + outputs[2:]  # add hidden states and attention if they are here

#         if labels is not None:
#             loss_fct = BCEWithLogitsLoss()  #改成多標籤分類(每個標籤做二元分類的感覺)
#             labels = labels.float()  
#             loss = loss_fct(logits.view(-1, self.num_labels), labels.view(-1, self.num_labels))
#             outputs = (loss,) + outputs 

#         return outputs  


class MyNewBertForTokenClassification(BertPreTrainedModel):
    '''將BERT每個位置輸出接上線性層預測屬性，並輔以屬性轉移矩陣設置；另外將BERT輸出額外接上另一線性層預測情感，之後用情感轉移矩陣獲得最終情感預測結果，是屬性情感分析的模型'''
    
    def __init__(self, config):
        super().__init__(config)

        self.bert = BertModel(config)
        self.dropout = nn.Dropout(config.hidden_dropout_prob)
        
        ### 將BERT輸出傳到這層來做屬性標記預測 ###
        self.aspect_classfier = nn.Linear(config.hidden_size,3)  #768 -> 3 (BIO) 
        
        
        ### 增加屬性轉移矩陣 ###
#         transition_path = {'B': ['B','I', 'O'],
#                                 'I': ['B', 'I', 'O'],
#                                 'O': ['B','O']}
        self.transition_scores_aspect = [[1/3,1/3,1/3],[1/3,1/3,1/3],[1/2,0,1/2]]  #3*3矩陣，為B/I/O到B/I/O的機率     
        self.transition_scores_aspect = np.array(self.transition_scores_aspect, dtype='float32').transpose()  
        
        
        ### 增加情感轉移矩陣 (FOR 屬性標籤預測情感標籤) ###
#         transition_path = {'B': ['NEG','NEU','POS'],
#                                 'I': ['NEG','NEU','POS'],
#                                 'O': ['O']}   
        self.transition_scores_sentiment = [[1/3,1/3,1/3,0],[1/3,1/3,1/3,0],[0,0,0,1]]  #3*4矩陣，為B/I/O到NEG/NEU/POS/O的機率     
        self.transition_scores_sentiment = np.array(self.transition_scores_sentiment, dtype='float32').transpose()  
          
        
        ### 將BERT輸出傳到這層來預測情感標記 ###
        self.sentiment_classifier = nn.Linear(config.hidden_size, 4) #768 -> 4 (POS/NEU/NEG/O)
        
        self.init_weights()

        
    def forward(
        self,
        input_ids=None,
        attention_mask=None,
        token_type_ids=None,
        position_ids=None,
        head_mask=None,
        inputs_embeds=None,
        labels=None,
    ):
        outputs = self.bert(
            input_ids,
            attention_mask=attention_mask,
            token_type_ids=token_type_ids,
            position_ids=position_ids,
            head_mask=head_mask,
            inputs_embeds=inputs_embeds,
        )
                 
        tagger_input = outputs[0]
        tagger_input = self.dropout(tagger_input)

        #### 以下開始對屬性部分做處理 ####
        aspect_tagger_pred = self.aspect_classfier(tagger_input)  
        
        pred_aspect = F.softmax(aspect_tagger_pred,dim=2)  
        
        if torch.cuda.is_available():
            self.transitions_aspect = torch.tensor(self.transition_scores_aspect).cuda() 
        else:
            self.transitions_aspect = torch.tensor(self.transition_scores_aspect) 
            
        combined_aspect_pred = []  
        
        for bsz in range(pred_aspect.size(0)) :  
            current_sample = pred_aspect[bsz]  
            
            current_sample_transition = []  
            for t in range(current_sample.size(0)):  
                if t == 0 or t == 1:  #對 cls & 對 cls後的第一個字 => 兩者都不用做轉移矩陣運算
                    current_sample_transition.append(current_sample[t]) 

                else:
                    transition_pred = torch.mm(self.transitions_aspect,current_sample_transition[t-1].view(3,-1))  #轉移矩陣乘上前一個token已經權衡相加後的值
                   
                    
                    ### 計算 confidence value (用轉移矩陣算出的值的重要程度) ###
                    val = torch.sum(current_sample_transition[t-1].view(3,-1) * current_sample_transition[t-1].view(3,-1)) 
                    alpha = val * 0.6  
                    new_aspect_pred = alpha * transition_pred + (1-alpha) * current_sample[t].view(3,-1) 
                    current_sample_transition.append(new_aspect_pred.view(-1))  

            combined_aspect_pred.append(torch.stack(current_sample_transition, 0))   
             
        aspect_tagger = torch.stack(combined_aspect_pred, 0)  

        
        ### 以下開始對情感部分做處理 ###
        sentiment_pred = self.sentiment_classifier(tagger_input) 
        
        if torch.cuda.is_available():
            self.transitions_sentiment = torch.tensor(self.transition_scores_sentiment).cuda()
        else:
            self.transitions_sentiment = torch.tensor(self.transition_scores_sentiment)
        
        sentiment_pred = F.softmax(sentiment_pred,dim=2)  
        
        combined_sentiment_pred = []  
        
        for bsz in range(sentiment_pred.size(0)) :  
            current_sample_aspect = aspect_tagger[bsz]  
            current_sample_sentiment = sentiment_pred[bsz] 
            
            current_sample_transition = []  
            for t in range(current_sample_sentiment.size(0)): 
                if t == 0:  #對cls不用做轉移矩陣運算
                    current_sample_transition.append(current_sample_sentiment[t])  
                else:
                    transition_pred = torch.mm(self.transitions_sentiment,current_sample_aspect[t].view(3,-1))  #轉移矩陣乘上屬性預測標籤
                   
                    ### 計算 confidence value (這個用轉移矩陣算出的值的重要程度) ###  
                    val = torch.sum(current_sample_aspect[t].view(3,-1) * current_sample_aspect[t].view(3,-1)) 
                    
                    alpha = val * 0.6 
                    
                    new_sentiment_pred = alpha * transition_pred + (1-alpha) * current_sample_sentiment[t].view(4,-1) 
                    current_sample_transition.append(new_sentiment_pred.view(-1))  

            combined_sentiment_pred.append(torch.stack(current_sample_transition, 0))    
             
        logits = torch.stack(combined_sentiment_pred, 0)  
        
        
        outputs = (aspect_tagger,) + (logits,) + outputs[2:]   
        
    
        ### 輸入的label map:  {'B-NEG': 0, 'B-NEU': 1, 'B-POS': 2, 'I-NEG': 3, 'I-NEU': 4, 'I-POS': 5, 'O': 6}
        ## 自定義的新label map : 
            # aspect_labels : {"B" : 0, "I":1, "O": 2}
            # sentiment_labels {"NEG": 0 ,"NEU":1, "POS":2, "O":3}
        
        if labels is not None:
            #處理 labels、讓它符合joint屬性模型格式 
            aspect_labels, sentiment_labels = chg_labels_to_aspect_and_sentiment(labels)

            aspect_tagger = torch.log(aspect_tagger + 1e-20)  #將屬性預測部分經過log
            logits = torch.log(logits + 1e-20)  #將情感預測部分經過log
            
            loss_fct = NLLLoss()  #因為已經過softmax和log，不能使用CrossEntropyLoss
            
            if attention_mask is not None: 
                active_loss = attention_mask.view(-1) == 1  
                active_logits = aspect_tagger.view(-1, 3)
                active_labels = torch.where(
                    active_loss, aspect_labels.view(-1), torch.tensor(loss_fct.ignore_index).type_as(labels)
                )
                aspect_loss = loss_fct(active_logits, active_labels)
                
                
                active_loss = attention_mask.view(-1) == 1  
                active_logits = logits.view(-1, 4)
                active_labels = torch.where(
                    active_loss, sentiment_labels.view(-1), torch.tensor(loss_fct.ignore_index).type_as(labels)
                )
                sentiment_loss = loss_fct(active_logits, active_labels)

                
                loss = aspect_loss + sentiment_loss   
                
            else:
                aspect_loss = loss_fct(aspect_tagger.view(-1, 3), aspect_labels.view(-1))
                sentiment_loss = loss_fct(logits.view(-1, 4), sentiment_labels.view(-1))
                
                loss = aspect_loss + sentiment_loss   
            
            outputs = (loss,) + (aspect_loss,) + (sentiment_loss,) + outputs  

        return outputs   


# class MyNewBertForTokenClassification(BertPreTrainedModel):
#     '''把bert輸出分別接上一層線性分類層，是屬性情感分析的模型'''
    
#     def __init__(self, config):
#         super().__init__(config)

#         self.bert = BertModel(config)
#         self.dropout = nn.Dropout(config.hidden_dropout_prob)
        
#         # 將bert輸出傳到這層來預測屬性標記
#         self.aspect_classfier = nn.Linear(config.hidden_size,3)  #768 -> 3 (BIO) 
        
#         # 將bert輸出傳到這層來預測情感標記
#         self.sentiment_classifier = nn.Linear(config.hidden_size, 4) #768 -> 4 (POS/NEU/NEG/O)
        
#         self.init_weights()

        
#     def forward(
#         self,
#         input_ids=None,
#         attention_mask=None,
#         token_type_ids=None,
#         position_ids=None,
#         head_mask=None,
#         inputs_embeds=None,
#         labels=None,
#     ):
#         outputs = self.bert(
#             input_ids,
#             attention_mask=attention_mask,
#             token_type_ids=token_type_ids,
#             position_ids=position_ids,
#             head_mask=head_mask,
#             inputs_embeds=inputs_embeds,
#         )
                 
#         tagger_input = outputs[0]
#         tagger_input = self.dropout(tagger_input)

#         aspect_tagger = self.aspect_classfier(tagger_input)  #輸出預測的屬性標註, 維度是bsz,seq,num_labels (3,因為標BIO) 
#         logits = self.sentiment_classifier(tagger_input)   #得到最後經過情感標註後的結果
        
#         outputs = (aspect_tagger,) + (logits,) + outputs[2:]   
    
    
#         ### 輸入的label map:  {'B-NEG': 0, 'B-NEU': 1, 'B-POS': 2, 'I-NEG': 3, 'I-NEU': 4, 'I-POS': 5, 'O': 6}
#         ## 需轉成自定義的新label map : 
#             # aspect_labels : {"B" : 0, "I":1, "O": 2}
#             # sentiment_labels {"NEG": 0 ,"NEU":1, "POS":2, "O":3}
        
#         if labels is not None:
#             #處理 labels、讓它符合joint屬性模型格式
#             aspect_labels, sentiment_labels = chg_labels_to_aspect_and_sentiment(labels)
             
#             loss_fct = CrossEntropyLoss()  
#             if attention_mask is not None: 
#                 # 先算屬性標記的loss:
#                 active_loss = attention_mask.view(-1) == 1  
#                 active_logits = aspect_tagger.view(-1, 3)
#                 active_labels = torch.where(
#                     active_loss, aspect_labels.view(-1), torch.tensor(loss_fct.ignore_index).type_as(labels)
#                 )
#                 aspect_loss = loss_fct(active_logits, active_labels)
                
                
#                 active_loss = attention_mask.view(-1) == 1  
#                 active_logits = logits.view(-1, 4)
#                 active_labels = torch.where(
#                     active_loss, sentiment_labels.view(-1), torch.tensor(loss_fct.ignore_index).type_as(labels)
#                 )
#                 sentiment_loss = loss_fct(active_logits, active_labels)
                
#                 loss = aspect_loss + sentiment_loss  #兩個loss會一起訓練
                
#             else:
#                 aspect_loss = loss_fct(aspect_tagger.view(-1, 3), aspect_labels.view(-1))
#                 sentiment_loss = loss_fct(logits.view(-1, 4), sentiment_labels.view(-1))
                
#                 loss = aspect_loss + sentiment_loss  #兩個loss會一起訓練
            
#             outputs = (loss,) + (aspect_loss,) + (sentiment_loss,) + outputs 

#         return outputs  